#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <time.h>

#include "gateway_cts.h"

int gateway_write_current_time(gatt_connection_t* connection)
{
    time_t ttime;
    current_time_t ct;
    struct tm* tmtime;

    uuid_t char_uuid;
	uint8_t status;

    /* Sanity Checks */
	if(NULL == connection)
	{
        fprintf(stderr, "Parameter \"gatt_connection_t* connection\" passed is NULL.\n");
        return GATTLIB_INVALID_PARAMETER;
	}

    /* Get the System Time */
    time(&ttime);

    /* Convert it to struct tm UTC time */
    tmtime = gmtime(&ttime);

    /* Populate the members of the current_time_t object */
    ct.year         = tmtime->tm_year + 1900;
    ct.month        = tmtime->tm_mon + 1;
    ct.day          = tmtime->tm_mday;
    ct.hours        = tmtime->tm_hour;
    ct.minutes      = tmtime->tm_min;
    ct.seconds      = (60 > tmtime->tm_sec)?tmtime->tm_sec:(0);
    ct.day_of_week  = (0 == tmtime->tm_wday)?(7):tmtime->tm_wday;
    ct.fractal256 = 0x00;
    ct.req_reason = 0x01;

    /* Convert the Current Time characteristic uuid from string to uuid_t */
	gattlib_string_to_uuid(CTS_CURRENT_TIME_CHARACTERISTIC_UUID, strlen(CTS_CURRENT_TIME_CHARACTERISTIC_UUID) + 1, &char_uuid);

	/* Attempt to write the contents of current_time_t object to the Current Time characteristics of CTS service */
	status = gattlib_write_char_by_uuid(connection, &char_uuid, &ct, sizeof(ct));

	/* If the operation failed print the error and return the error code */
	if (status != GATTLIB_SUCCESS)
	{
		char uuid_str[MAX_LEN_UUID_STR + 1];
		gattlib_uuid_to_string(&char_uuid, uuid_str, sizeof(uuid_str));

		if (status == GATTLIB_NOT_FOUND)
		{
			fprintf(stderr, "Could not find GATT Characteristic (CURRENT_TIME) with UUID %s.\n", uuid_str);
		}
		else
		{
			fprintf(stderr, "Error while writing GATT Characteristic (CURRENT_TIME) with UUID %s (ret:%d)\n", uuid_str, status);
		}
	}
	return status;

}

int gateway_request_current_time(gatt_connection_t* connection, current_time_t** read_time)
{
	uint8_t* cts_buffer = NULL;
	uint8_t status;
	size_t bytes_read;
	uuid_t char_uuid;
	uint8_t i;

    /* Sanity Checks */
	if(NULL == connection)
	{
        fprintf(stderr, "Parameter \"gatt_connection_t* connection\" passed is NULL.\n");
        return GATTLIB_INVALID_PARAMETER;
	}
    if(NULL == read_time)
	{
        fprintf(stderr, "Parameter \"current_time_t** read_time\" passed is NULL.\n");
        return GATTLIB_INVALID_PARAMETER;
	}

	/* Convert the Current Time characteristic uuid from string to uuid_t */
	gattlib_string_to_uuid(CTS_CURRENT_TIME_CHARACTERISTIC_UUID, strlen(CTS_CURRENT_TIME_CHARACTERISTIC_UUID) + 1, &char_uuid);

	/*
	 * Attempt to read the value of Current Time characteristic of CTS service from the remote device through the passed pre-established connection.
	 * Space will be automatically allocated if anything is returned with length larger than zero and it will be the responsibility of the caller to free it.
	 */
	status = gattlib_read_char_by_uuid(connection, &char_uuid, (void **)&cts_buffer, &bytes_read);
	if (status != GATTLIB_SUCCESS)
	{
		char uuid_str[MAX_LEN_UUID_STR + 1];

		gattlib_uuid_to_string(&char_uuid, uuid_str, sizeof(uuid_str));
		if (status == GATTLIB_NOT_FOUND)
		{
			fprintf(stderr, "Could not find Characteristic (CURRENT_TIME) with UUID %s.\n", uuid_str);
		}
		else
		{
			fprintf(stderr, "Error while reading GATT Characteristic (CURRENT_TIME) with UUID %s (ret:%d)\n", uuid_str, status);
		}
		return status;
	}
	else
	{
        if(sizeof(current_time_t) == bytes_read)
        {
            printf("\nYear\tMonth\tDay\tHour\tMin\tSec\tDotW\t256F\tReason\t\n");

            /* Convert the bytes representing the Year into uint16_t before printing it out */
            printf("%d\t", ((cts_buffer[1] << 8) | cts_buffer[0]));

            /* Print out all the other Bytes representing Current Time */
            for (i = 2; i < 10; i++)
            {
                printf("%d\t", cts_buffer[i]);
            }
            printf("\n\n");

            /* Pass the pointer to the allocated space to the caller. Make sure to free it from the caller's side */
            *read_time = (current_time_t*)cts_buffer;
        }
    }

    return status;
}
