
#ifndef GATEWAY_CTS_
#define GATEWAY_CTS_

#include "gattlib.h"

#define CTS_CURRENT_TIME_CHARACTERISTIC_UUID			"2A2B"

/* ============================== Custom Type Definitions  ============================== */

/*
 * Struct to hold the current time. Reflects the format that is implemented on the remote device.
 */
typedef struct __attribute__((packed,aligned(1)))
{
    uint16_t    year;                       // 0 - Unknown, Range 1582 - 9999
    uint8_t     month;                      // 1-12, 1 - January, ... , 12 - December
    uint8_t     day;                        // 1-31
    uint8_t     hours;                      // 0-23
    uint8_t     minutes;                    // 0-59
    uint8_t     seconds;                    // 0-59
    uint8_t     day_of_week;                // 0 - Unknown, 1 - Monday, ... , 7 - Sunday
    uint8_t     fractal256;                 // 1/256 of a second. In most cases the required accuracy is second based so typically set to 0.
    uint8_t     req_reason;                 // When setting the remote device time, use value 0x01.
} current_time_t;


/* ================================= Function Prototypes ================================ */

/**
 * @brief This function sets the current_time (GMT time) into Current Time Service of the device.
 *
 * The function sets the fields of a current_time_t object and then writes the contents of the object
 * to the Current Time characteristic of Current Time Service(CTS) specified by SIG. Characteristic UUID is 2A2B.
 *
 * @param[in]   connection  gatt_connection_t type pointer pointing to the current connection.
 *
 * @return      status      uint8_t value that represents an error code as defined in gattlib.h header file.
 */
int gateway_write_current_time(gatt_connection_t* connection);



/**
 * @brief This function requests the Current Time charaacteristic value of Current Time Service of the remote device.
 *
 * The function reads Current Time characteristic of Current Time Service(CTS) as specified by SIG with UUID is 2A2B.
 * Once the characteristic is read, the read value is printed on the screen and the pointer pointing to the read value will
 * be passed to the caller if a complete current_time_t structure instance was read.
 *
 * DO NOT allocate memory in the caller as space will be automatically allocated when calling gateway_request_current_time() function
 * as long as a valid complete current_time_t object was read succesfully. The responsibility to free() the allocated space is passed
 * to the caller of this API.
 *
 * @param[in]   connection  gatt_connection_t type pointer pointing to the current connection.
 * @param[out]  read_time   pointer to current_time_t type pointer which will contain the Current Time characteristic value read.
 *
 * @return      status      uint8_t value that represents an error code as defined in gattlib.h header file.
 */
int gateway_request_current_time(gatt_connection_t* connection, current_time_t** read_time);

#endif /* GATEWAY_CTS_ */
