	#include <assert.h>
	#include <stdio.h>
	#include <stdlib.h>
	#include <string.h>

	#include "gateway_logs.h"

/* ================================= Request Device Info ================================ */

int gateway_request_device_info(gatt_connection_t* connection, id_info_t** dev_info)
{
	uint8_t status;
	size_t bytes_read;
	uuid_t char_uuid;

    /* Sanity Checks */
	if(NULL == connection)
	{
        fprintf(stderr, "Parameter \"gatt_connection_t* connection\" passed is NULL.\n");
        return GATTLIB_INVALID_PARAMETER;
	}
    if(NULL == dev_info)
	{
        fprintf(stderr, "Parameter \"d_info_t** dev_info\" passed is NULL.\n");
        return GATTLIB_INVALID_PARAMETER;
	}

	/* Convert the Current Time characteristic uuid from string to uuid_t */
	gattlib_string_to_uuid(DEVICE_INFO_CHARACTERISTIC_UUID, strlen(DEVICE_INFO_CHARACTERISTIC_UUID) + 1, &char_uuid);

	/*
	 * Attempt to read the value of Current Time characteristic of CTS service from the remote device through the passed pre-established connection.
	 * Space will be automatically allocated if anything is returned with length larger than zero and it will be the responsibility of the caller to free it.
	 */
	status = gattlib_read_char_by_uuid(connection, &char_uuid, (void**)dev_info, &bytes_read);
	if (status != GATTLIB_SUCCESS)
	{
		char uuid_str[MAX_LEN_UUID_STR + 1];

		gattlib_uuid_to_string(&char_uuid, uuid_str, sizeof(uuid_str));
		if (status == GATTLIB_NOT_FOUND)
		{
			fprintf(stderr, "Could not find Characteristic (REMOTE_READ_RSSI - DEVICE_INFO) with UUID %s.\n", uuid_str);
		}
		else
		{
			fprintf(stderr, "Error while reading GATT Characteristic (REMOTE_READ_RSSI - DEVICE_INFO) with UUID %s (ret:%d)\n", uuid_str, status);
		}
		return status;
	}


    return status;
}

/* ================================= Request Zone X Logs ================================ */

/**
 * @brief This function sets the value of characteristic ZONE in order to later request the logs from the remote device for that particular zone.
 *
 * @param[in]   connection  gatt_connection_t type pointer pointing to the current connection.
 * @param[in]   zone        uint8_t value representing the requested danger zone taking values from 0 to 3.
 *
 * @return      status      int value that represents an error code as defined in gattlib.h header file.
 */
static int gateway_remote_device_zone_set(gatt_connection_t* connection, uint8_t zone)
{
	uuid_t char_uuid;
	uint8_t status;

	/* Sanity Checks */
	if(NULL == connection)
	{
        fprintf(stderr, "Parameter \"gatt_connection_t* connection\" passed is NULL.\n");
        return GATTLIB_INVALID_PARAMETER;
	}

	/* Convert the ZONE characteristic UUID from const string to uuid_t */
	gattlib_string_to_uuid(ZONE_CHARACTERISTIC_UUID, strlen(ZONE_CHARACTERISTIC_UUID) + 1, &char_uuid);

	/* Attempt to write to ZONE characteristic the value passed as parameter zone */
	status = gattlib_write_char_by_uuid(connection, &char_uuid, &zone, sizeof(zone));
	if (status != GATTLIB_SUCCESS)
	{
		char uuid_str[MAX_LEN_UUID_STR + 1];
		gattlib_uuid_to_string(&char_uuid, uuid_str, sizeof(uuid_str));

		if (status == GATTLIB_NOT_FOUND)
		{
			fprintf(stderr, "Could not find GATT Characteristic (ZONE) with UUID %s.\n", uuid_str);
		}
		else
		{
			fprintf(stderr, "Error while writing GATT Characteristic (ZONE) with UUID %s (ret:%d)\n", uuid_str, status);
		}
	}
	return status;

}

int gateway_request_zone_x_log_dump(gatt_connection_t* connection, FILE* fptr, uint8_t zone)
{
	uint8_t* log_buffer = NULL;
	uint8_t status;
	size_t bytes_read;
	uuid_t char_uuid;

	/* Sanity Checks */
	if(NULL == connection)
	{
        fprintf(stderr, "Parameter \"gatt_connection_t* connection\" passed is NULL.\n");
        return GATTLIB_INVALID_PARAMETER;
	}
	if(NULL == fptr)
	{
        fprintf(stderr, "Parameter \"FILE* fptr\" passed is NULL.\n");
        return GATTLIB_INVALID_PARAMETER;
	}

    /* Attempt to set ZONE characteristic to the zone that the logs of are to be requested */
	status = gateway_remote_device_zone_set(connection, zone);
	if(status != GATTLIB_SUCCESS)
    {
        printf("Failed to set ZONE characteristic to value %d. Status = %d", zone, status);
        return status;
    }

    /* Convert the ZONE_X_LOG characteristic UUID from const string to uuid_t */
	gattlib_string_to_uuid(ZONE_X_LOG_CHARACTERISTIC_UUID, strlen(ZONE_X_LOG_CHARACTERISTIC_UUID) + 1, &char_uuid);

	/*
	 * While no zero length result is returned, repeatedly request for ZONE_X_LOG characteristic value. This way it is ensured that the
	 * entirety of the logs of the zone will be retrieved.
	 */
	do
	{
        /*
         * Attempts to read the value of ZONE_X_LOG characteristic. Read automatically allocates memory for the buffer if the length of the
         * read value is larger than zero and returns a pointer to it. It is the responsibility of the caller to free it when it is no longer
         * needed. Do not forget to free() once done with the data.
         */
		status = gattlib_read_char_by_uuid(connection, &char_uuid, (void **)&log_buffer, &bytes_read);
		if (status != GATTLIB_SUCCESS)
		{
			char uuid_str[MAX_LEN_UUID_STR + 1];

			gattlib_uuid_to_string(&char_uuid, uuid_str, sizeof(uuid_str));

			if (status == GATTLIB_NOT_FOUND) {
				fprintf(stderr, "Could not find GATT Characteristic (ZONE_X_LOG) with UUID %s.\n", uuid_str);
			}
			else
			{
				fprintf(stderr, "Error while reading GATT Characteristic (ZONE_X_LOG) with UUID %s (ret:%d)\n", uuid_str, status);
			}
			/* If an error of the above kinds has occured there will be no memory allocation. No need to free. */
			return status;
		}
		else
		{
            /* Write bytes equal to 'bytes_read' from the log_buffer to the file pointed to by FILE type pointer parameter fptr. */
            fwrite(log_buffer, 1, bytes_read, fptr);
		}

        /* Free the space allocated to the buffer if a non zero length value was read otherwise there has been no memory allocation */
        if(0 != bytes_read)
        {
            free(log_buffer);
        }

	} while (bytes_read != 0);

	return status;
}


/* ================================== Reset Log Index  ================================== */

int gateway_reset_zone_index(gatt_connection_t* connection, int8_t zone)
{
	uuid_t char_uuid;
	uint8_t status;

    /* Sanity Checks */
	if(NULL == connection)
	{
        fprintf(stderr, "Parameter \"gatt_connection_t* connection\" passed is NULL.\n");
        return GATTLIB_INVALID_PARAMETER;
	}

	/* Convert the PTR_CONFIG characteristic UUID from const string to uuid_t */
	gattlib_string_to_uuid(PTR_CONFIG_CHARACTERISTIC_UUID, strlen(PTR_CONFIG_CHARACTERISTIC_UUID) + 1, &char_uuid);

	/* Attempt to write to PTR_CONFIG characteristic the value of zone passed as parameter */
	status = gattlib_write_char_by_uuid(connection, &char_uuid, &zone, sizeof(zone));
			if (status != GATTLIB_SUCCESS)
			{
				char uuid_str[MAX_LEN_UUID_STR + 1];

				gattlib_uuid_to_string(&char_uuid, uuid_str, sizeof(uuid_str));

				if (status == GATTLIB_NOT_FOUND)
				{
					fprintf(stderr, "Could not find GATT Characteristic (PTR_CONFIG) with UUID %s.\n", uuid_str);
				}
				else
				{
					fprintf(stderr, "Error while writing GATT Characteristic (PTR_CONFIG) with UUID %s (ret:%d)\n",
						uuid_str, status);
				}
			}

	return status;
}

int gateway_reset_all_indices(gatt_connection_t* connection)
{
	int8_t i;
	uint8_t status;

    /* Sanity Checks */
	if(NULL == connection)
	{
        fprintf(stderr, "Parameter \"gatt_connection_t* connection\" passed is NULL.\n");
        return GATTLIB_INVALID_PARAMETER;
	}

	for(i = 3; i > -1; i--)
	{
		status = gateway_reset_zone_index(connection, i);
		if(GATTLIB_SUCCESS != status)
		{
			printf("Failed to reset index of Zone %d", i);
			return status;
		}
	}
	return status;
}

/* ================================== Truncate Folder  ================================== */

int gateway_truncate_zone_folder(gatt_connection_t* connection, int8_t zone)
{
	uuid_t char_uuid;
	uint8_t status;

    /* Sanity Checks */
	if(NULL == connection)
	{
        fprintf(stderr, "Parameter \"gatt_connection_t* connection\" passed is NULL.\n");
        return GATTLIB_INVALID_PARAMETER;
	}

	/* Convert the FLD_RESET characteristic UUID from const string to uuid_t */
	gattlib_string_to_uuid(FLD_RESET_CHARACTERISTIC_UUID, strlen(FLD_RESET_CHARACTERISTIC_UUID) + 1, &char_uuid);

	/* Attempt to write to FLD_RESET characteristic the value of zone passed as parameter */
	status = gattlib_write_char_by_uuid(connection, &char_uuid, &zone, sizeof(zone));
			if (status != GATTLIB_SUCCESS)
			{
				char uuid_str[MAX_LEN_UUID_STR + 1];

				gattlib_uuid_to_string(&char_uuid, uuid_str, sizeof(uuid_str));

				if (status == GATTLIB_NOT_FOUND)
				{
					fprintf(stderr, "Could not find GATT Characteristic (FLD_RESET) with UUID %s. "
						"You might call the program with '--gatt-discovery'.\n", uuid_str);
				}
				else
				{
					fprintf(stderr, "Error while writing GATT Characteristic (FLD_RESET) with UUID %s (ret:%d)\n",
						uuid_str, status);
				}
			}

	return status;
}

int gateway_truncate_all_zone_folders(gatt_connection_t* connection)
{
	int8_t i;
	uint8_t status;

    /* Sanity Checks */
	if(NULL == connection)
	{
        fprintf(stderr, "Parameter \"gatt_connection_t* connection\" passed is NULL.\n");
        return GATTLIB_INVALID_PARAMETER;
	}

	for(i = 3; i > -1; i--)
	{
		status = gateway_truncate_zone_folder(connection, i);
		if(GATTLIB_SUCCESS != status)
		{
			printf("Failed to truncate Zone %d folder.\n", i);
			return status;
		}
	}
	return status;
}
