
#ifndef GATEWAY_LOGS_
#define GATEWAY_LOGS_

#include "gattlib.h"

/* Size of ZONE_X_LOG characteristic. */
#define ZONE_X_LOG_CHARACTERISTIC_SIZE                      512

/* Characteristic UUID definitions */
#define ZONE_X_LOG_CHARACTERISTIC_UUID  					"3112fb86-86c7-11ea-bc55-000000000003"
#define ZONE_CHARACTERISTIC_UUID    						"3112fb86-86c7-11ea-bc55-000000000004"
#define PTR_CONFIG_CHARACTERISTIC_UUID						"3112fb86-86c7-11ea-bc55-000000000005"
#define FLD_RESET_CHARACTERISTIC_UUID						"3112fb86-86c7-11ea-bc55-000000000006"

/* The following UUID corresponds to REMOTE_READ_RSSI characteristic which when read, returns device information */
#define DEVICE_INFO_CHARACTERISTIC_UUID                     "8522c64b-ed9d-4789-a95d-000000000001"

/* Number of bytes per row of printed/written_to_file data */
#define BYTES_PER_ROW   									(20)

/* ============================== Custom Type Definitions  ============================== */

/*
 * Struct to hold the device information. Reflects the format that is implemented on the remote device for struct id_info.
 */
typedef struct __attribute__((packed,aligned(1)))
{
    uint32_t    DevUID;                         // Device Uinique Identifier
    uint8_t     DevType;                        // Device Type
    uint8_t     DevToF;                         // Is the device capable of performing the Time of Flight process ? 0 - False 1 - True
    uint8_t     Dev_conn_to_me;                 // Should the SDT connect to this particular device or is it simply a beacon ? 0 - False 1 - True
} id_info_t;

/* ================================= Function Prototypes ================================ */

/* ================================= Request Device Info ================================ */

/**
 * @brief This function requests the device information of the remote device.
 *
 * The function reads REMOTE_READ_RSSI characteristic with UUID 8522c64b-ed9d-4789-a95d-000000000001.
 * Once the characteristic is read, the callback triggered on the remote device will return sizeof(id_info) bytes.
 * Refer to typedef struct id_info definition in the remote device firmware for a better understanding of the format of the received data.
 *
 * The structure is packed to ensure that no padding is added so sizeof(d_info_t) will return 7 bytes instead of 8.
 *
 * DO NOT allocate memory in the caller as space will be automatically allocated when calling gateway_request_device_info() function
 * as long as a valid complete d_info_t object was succesfully read. The responsibility to free() the allocated space is passed
 * to the caller of this API.
 *
 * @param[in]   connection  gatt_connection_t type pointer pointing to the current connection.
 * @param[out]  dev_info    pointer to d_info_t type pointer which hosts the device information read.
 *
 * @return      status      int8_t value that represents an error code as defined in gattlib.h header file.
 */
int gateway_request_device_info(gatt_connection_t* connection, id_info_t** dev_info);

/* ================================= Request Zone X Logs ================================ */

/**
 * @brief This function requests logs of a zone from the remote device, saves them on the file and prints them to the console.
 *
 * The function requires an existing connection and an open file pointed to by FILE type pointer fptr. First, it will call  remote_device_zone_set()
 * function to setup the ZONE characteristic to the value of the zone parameter.
 *
 * Once the ZONE characteristic is successfully set, the gateway will repeatedly request to read the value of characteristic ZONE_X_LOG
 * until a zero length read value is returned. For each successful read of the characteristic, the contents of the buffer are then printed on the
 * screen into rows of BYTES_PER_ROW Bytes.
 *
 * In a similar way, the contents are also written to the file pointed to by FILE type pointer fptr.
 *
 * Finally the function returns an integer value that represents the error code that might have occured or GATTLIB_SUCCESS (value 0) if the process
 * completed successfully.
 *
 * @param[in]   connection  gatt_connection_t type pointer pointing to the current connection.
 * @param[out]  fptr        pointer to the opened file that the logs will be stored.
 * @param[in]   zone        uint8_t value representing the requested danger zone taking values from 0 to 3.
 *
 * @return      status      int value that represents an error code as defined in gattlib.h header file.
 */
int gateway_request_zone_x_log_dump(gatt_connection_t* connection, FILE* fptr, uint8_t zone);

/* ================================== Reset Log Index  ================================== */

/**
 * @brief This function updates characteristic PTR_CONFIG.
 *
 * The function will attempt to write to the characteristic PTR_CONFIG the value passed as parameter zone. This will signify to the remote device,
 * that the file index of its saved logs needs to be reset so the logs can be read from the beginning. Please note that truncating the folder will
 * automatically reset its index as well.
 *
 * @param[in]   connection  gatt_connection_t type pointer pointing to the current connection.
 * @param[in]   zone        uint8_t value representing the requested danger zone taking values from 0 to 3.
 *
 * @return      status      int value that represents an error code as defined in gattlib.h header file.
 */
int gateway_reset_zone_index(gatt_connection_t* connection, int8_t zone);

/**
 * @brief This function will attempt to reset all log indices.
 *
 * @param[in]   connection  gatt_connection_t type pointer pointing to the current connection.
 *
 * @return      status      int value that represents an error code as defined in gattlib.h header file.
 */
int gateway_reset_all_indices(gatt_connection_t* connection);

/* ================================== Truncate Folder  ================================== */
/**
 * @brief This function updates characteristic FLD_RESET.
 *
 * The function will attempt to write to the characteristic FLD_RESET, the value passed as parameter zone. This will signify to the remote device,
 * that the Log Folder corresponding to the specified zone needs to be truncated. Please note that truncating the folder will automatically reset
 * its index as well.
 *
 * @param[in]   connection  gatt_connection_t type pointer pointing to the current connection.
 * @param[in]   zone        uint8_t value representing the requested danger zone taking values from 0 to 3.
 *
 * @return      status      int value that represents an error code as defined in gattlib.h header file.
 */
int gateway_truncate_zone_folder(gatt_connection_t* connection, int8_t zone);

/**
 * @brief This function will attempt to truncate all the Log Folders one by one(see gateway_truncate_zone_folder() function).
 *
 * @param[in]   connection  gatt_connection_t type pointer pointing to the current connection.
 *
 * @return      status      int value that represents an error code as defined in gattlib.h header file.
 */
int gateway_truncate_all_zone_folders(gatt_connection_t* connection);

#endif /* GATEWAY_LOGS_ */
