#include <pthread.h>
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/queue.h>

#include "gattlib.h"
#include "gateway_cts.h"
#include "gateway_logs.h"

#define BLE_SCAN_TIMEOUT   4

#define UUID_FILTER_VALUE               "d643d847-157f-07a1-e741-6fef4d79e9c6"
#define PATH_AND_FILENAME               "/home/pi/Desktop/logdev_0x"

typedef void (*ble_discovered_device_t)(const char* addr, const char* name);

// We use a mutex to make the BLE connections synchronous
static pthread_mutex_t g_mutex = PTHREAD_MUTEX_INITIALIZER;

LIST_HEAD(listhead, connection_t) g_ble_connections;
struct connection_t {
	pthread_t thread;
	char* addr;
	LIST_ENTRY(connection_t) entries;
};

static void *ble_connect_device(void *arg) {
	struct connection_t *connection = arg;
	char* addr = connection->addr;
	gatt_connection_t* gatt_connection;
	FILE* fptr;
	current_time_t* ct;
	id_info_t* id_info;
	char complete_fpath[50];
	char suffix[10];
	int ret;

    /* Acquire the MUTEX */
	pthread_mutex_lock(&g_mutex);

	printf("------------START %s ---------------\n", addr);

	/* Connect to the device with addr as BDAddress */
	gatt_connection = gattlib_connect(NULL, addr, GATTLIB_CONNECTION_OPTIONS_LEGACY_DEFAULT);
	if (gatt_connection == NULL) {
		fprintf(stderr, "Fail to connect to the bluetooth device.\n");
		goto connection_exit;
	} else {
		puts("Succeeded to connect to the bluetooth device.");
	}

	/* Request to read the Current Time characteristic value of SIG defined service Current Time Service - for demonstration purposes. */
    ret = gateway_request_current_time(gatt_connection, &ct);
	if (ret != GATTLIB_SUCCESS) {
		fprintf(stderr, "Failed to read the Current Time of the remote device.\n");
		goto disconnect_exit;
	}

	/* Remember to free the memory allocated by calling gateway_request_current_time */
	free(ct);

	/* Request to set the Current Time characteristic value using the system clock(GMT Time) */
	ret = gateway_write_current_time(gatt_connection);
	if (ret != 0) {
		fprintf(stderr, "Failed to set the Current Time of the remote device.\n");
		goto disconnect_exit;
	}

	/* Request to read the Current Time characteristic value of SIG defined service Current Time Service - for verification purposes. */
    ret = gateway_request_current_time(gatt_connection, &ct);
	if (ret != 0) {
		fprintf(stderr, "Failed to read the Current Time of the remote device.\n");
		goto disconnect_exit;
	}

	/* Remember to free the memory allocated by calling gateway_request_current_time */
	free(ct);


    /* Reset all File indices as it is impossible to know if the previous device that requested to Read Logs crashed */
    ret = gateway_reset_all_indices(gatt_connection);
	if (ret != 0) {
		fprintf(stderr, "Fail to reset the indices of the remote device.\n");
		goto disconnect_exit;
	}

	/* Request to read the device information in order to use its unique id to form the filename of the file that will host the logs. */
	ret = gateway_request_device_info(gatt_connection, &id_info);
    if (ret != 0) {
        fprintf(stderr, "Fail to reset the indices of the remote device.\n");
        goto disconnect_exit;
	}

	/* Form the desired filename of the file that will host the logs of this device. */
	sprintf(suffix, "%x", id_info->DevUID);
	strcpy(complete_fpath, PATH_AND_FILENAME);
	strcat(complete_fpath, suffix);
	strcat(complete_fpath, ".log");

	/* Remember to free the memory allocated internally by gateway_request_device_info() function */
	free(id_info);

	/* Open the file using the formed filename, for write/read binary and FAIL if the file is not opened */
	fptr = fopen(complete_fpath,"wb+");
	if(NULL == fptr)
	{
		fprintf(stderr, "Unable to open file in path: %s.\n", complete_fpath);
		goto disconnect_exit;
	}


    /* Request to read the logs of all the danger zones and save them in the opened file */
	for(int i = 3; i >= 0; i--)
	{
        ret = gateway_request_zone_x_log_dump(gatt_connection, fptr, i);
        if (ret != 0)
        {
            fprintf(stderr, "Fail to read the logs of Danger Zone %d.\n", i);
            goto disconnect_exit;
        }
	}


    /* Make sure to truncate the folders of the remote device that correspond to each zone to free some space on the Flash */
	ret = gateway_truncate_all_zone_folders(gatt_connection);
    if (ret != 0) {
        fprintf(stderr, "Fail to reset the indices of the remote device.\n");
        goto disconnect_exit;
    }

    /* Disconnect to free the adapter resources */
disconnect_exit:
	gattlib_disconnect(gatt_connection);

    /* Close the file if one was opened */
	if(NULL != fptr)
	{
        fclose(fptr);
	}

connection_exit:
	printf("------------DONE %s ---------------\n", addr);

	/* Return the MUTEX so another thread can take it */
	pthread_mutex_unlock(&g_mutex);

	/* Return */
	return NULL;
}

/* Run for each device discovered */
static void ble_discovered_device(void *adapter, const char* addr, const char* name, void *user_data) {
	struct connection_t *connection;
	int ret;

	/* Display information about the discovered device to the console */
	if (name) {
		printf("Discovered %s - '%s'\n", addr, name);
	} else {
		printf("Discovered %s\n", addr);
	}

    /* Properly allocate memory for the connection that will follow */
	connection = malloc(sizeof(struct connection_t));
	if (connection == NULL) {
		fprintf(stderr, "Failt to allocate connection.\n");
		return;
	}

	/* Set the BDAddress of the device to connect to to the value of addr parameter */
	connection->addr = strdup(addr);

	/*
	 * Create a thread, store its handle in &connection->thread, using the default parameters. The
	 * thread will execute the code of ble_connect_device with passed argument a pointer to the
	 * connection structure and insert the connection to the list.
	 */
	ret = pthread_create(&connection->thread, NULL,	ble_connect_device, connection);
	if (ret != 0) {
		fprintf(stderr, "Failt to create BLE connection thread.\n");
		free(connection);
		return;
	}
	LIST_INSERT_HEAD(&g_ble_connections, connection, entries);
}

int main(int argc, const char *argv[]) {

    uuid_t f_uuid;
    uuid_t* uuid_filter[]={&f_uuid,NULL};
    const char* adapter_name;
	void* adapter;
	int ret;

	/* Make a uuid_t type object from string in order to use as filter when scanning for valid devices to connect to */
	gattlib_string_to_uuid(UUID_FILTER_VALUE, strlen(UUID_FILTER_VALUE) + 1, &f_uuid);

	/* Passed Arguments Handling */
	if (argc == 1) {
		adapter_name = NULL;
	} else if (argc == 2) {
		adapter_name = argv[1];
	} else {
		fprintf(stderr, "%s [<bluetooth-adapter>]\n", argv[0]);
		return 1;
	}

	/* Initialize a List that will hold the devices to connect to */
	LIST_INIT(&g_ble_connections);

	/* Open the BLE adapter */
	ret = gattlib_adapter_open(adapter_name, &adapter);
	if (ret) {
		fprintf(stderr, "ERROR: Failed to open adapter.\n");
		return 1;
	}

	/* Acquire the MUTEX */
	pthread_mutex_lock(&g_mutex);

	/* Scan for devices that contain any of the UUIDs in the uuid_filter array as part of their service */
	gattlib_adapter_scan_enable_with_filter(adapter, uuid_filter, 0, GATTLIB_DISCOVER_FILTER_USE_UUID, ble_discovered_device, BLE_SCAN_TIMEOUT, NULL);
	if (ret) {
		fprintf(stderr, "ERROR: Failed to scan.\n");
		goto EXIT;
	}

	/* Stop scanning */
	gattlib_adapter_scan_disable(adapter);

	puts("Scan completed");

	/* Release the MUTEX */
	pthread_mutex_unlock(&g_mutex);

	/* Wait for the threads to complete and join them */
	while (g_ble_connections.lh_first != NULL) {
		struct connection_t* connection = g_ble_connections.lh_first;
		pthread_join(connection->thread, NULL);
		LIST_REMOVE(g_ble_connections.lh_first, entries);
		free(connection->addr);
		free(connection);
	}


EXIT:
    /* Close the adapter and exit */
	gattlib_adapter_close(adapter);
	return ret;
}
