/**
 ****************************************************************************************
 *
 * @file bas_service.c
 *
 * @brief Battery service implementation
 *
 * Copyright (C) 2015-2020 Dialog Semiconductor.
 * This computer program includes Confidential, Proprietary Information
 * of Dialog Semiconductor. All Rights Reserved.
 *
 ****************************************************************************************
 */

/*
 * Current Time Service data
 */
#if (SDT_BATTERY_MONITORING_ENABLE == 1)

#include <stdio.h>
#include <stdbool.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include "osal.h"
#include "sys_watchdog.h"
#include "sys_power_mgr.h"
#include "ble_att.h"
#include "ble_common.h"
#include "ble_gap.h"
#include "ble_gatts.h"
#include "ble_service.h"
#include "ble_uuid.h"
#include "bas.h"
#include "commons.h"
#include "ad_gpadc.h"
#include "platform_devices.h"

/* Battery Service instance */
__RETAINED ble_service_t *bas;

/* Timer used for battery monitoring */
__RETAINED OS_TIMER bas_tim;

/*
 * The values depend on the battery type.
 * MIN_BATTERY_LEVEL (in mVolts) must correspond to dg_configBATTERY_LOW_LEVEL (in ADC units).
 */
#define MAX_BATTERY_LEVEL    4200
#define MIN_BATTERY_LEVEL    2800

/* Battery timer callback functions */
void bas_tim_cb(OS_TIMER timer)
{
        OS_TASK task = (OS_TASK) OS_TIMER_GET_TIMER_ID(timer);

        OS_TASK_NOTIFY(task, BAS_TMO_NOTIF, OS_NOTIFY_SET_BITS);
}

/* Convert voltage level to percentage */
static uint8_t bat_level(uint16_t voltage)
{
        if (voltage >= MAX_BATTERY_LEVEL) {
                return 100;
        } else if (voltage <= MIN_BATTERY_LEVEL) {
                return 0;
        }

        /*
         * Currently, discharging (Volt vs Capacity) is approximated by a linear function.
         * The exact formula depends on the specific battery being used.
         * A lookup table with the battery characterization should be created and used for
         * accurate battery charge indication
         */
        return (uint8_t) ((int) (voltage - MIN_BATTERY_LEVEL) * 100 /
                                                        (MAX_BATTERY_LEVEL - MIN_BATTERY_LEVEL));
}

/* Read the battery level by invoking the GPADC */
uint8_t read_battery_level(void)
{
        uint16_t bat_voltage;
        uint16_t value;
        ad_gpadc_handle_t handle;

        handle = ad_gpadc_open(&BATTERY_LEVEL);
        ad_gpadc_read(handle, &value);
        bat_voltage = ad_gpadc_conv_to_batt_mvolt(BATTERY_LEVEL.drv, value);
        ad_gpadc_close(handle, false);

        printf("BAT voltage: %dmV\r\n", bat_voltage);

        return bat_level(bat_voltage);
}

/* Update battery level */
void bas_update(bool sdt_charging)
{
        uint8_t level;

        level = read_battery_level();

        printf("BAT: %d%%\r\n", level);

        bas_set_level(bas, level, false);

        if (level <= 1) {
                if (!sdt_charging) {
                        printf("BAT-LOW\r\n");
                        pm_sleep_mode_set(pm_mode_hibernation);
                }
        }
}
#endif /* SDT_BATTERY_MONITORING_ENABLE == 1 */
