/**
 ****************************************************************************************
 *
 * @file distance_sw_fsm.c
 *
 * @brief Social distance tag application
 *
 * Copyright (C) 2015-2020 Dialog Semiconductor.
 * This computer program includes Confidential, Proprietary Information
 * of Dialog Semiconductor. All Rights Reserved.
 *
 ****************************************************************************************
 */

#include <stdbool.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <stdio.h>
#include "sdk_queue.h"
#include "osal.h"
#include "sys_watchdog.h"
#include "ble_att.h"
#include "ble_common.h"
#include "ble_gap.h"
#include "ble_gatts.h"
#include "ble_gattc.h"
#include "ble_service.h"
#include "ble_uuid.h"
#include "cts.h"
#include "bas.h"
#include "ad_nvms.h"
#include "sys_trng.h"
#include "timers.h"
#include "hw_timer.h"
#include "att.h"
#include "ble_gattc_util.h"
#include "sys_clock_mgr.h"
#include "hw_rtc.h"
#include "sys_power_mgr.h"
#include "time_service.h"
#include "ble_custom_service.h"
#include "commons.h"
#include "ad_nvparam.h"
#include "platform_nvparam.h"
#include "distance_sw_fsm.h"

#if ( SDT_TOF_RANGING_ENABLE == 1 )
#include "dte_common.h"
#if ENABLE_CW_DIST_CALC
        #include "cmac_iq_sampling.h"
        #include "cw_distance.h"
#endif /* ENABLE_CW_DIST_CALC */
#endif /* SDT_TOF_RANGING_ENABLE == 1 */

#if ( SDT_LOGGING_ENABLE == 1 )
        #include "miscellaneous.h"
#endif

#if ( SDT_BATTERY_MONITORING_ENABLE == 1 )
        #include "bas_service.h"
#endif


/*********************************************************************
 *
 *      Defines
 */

/* Define the size of the RSSI characteristic to 1 signed int8_t size*/
#define APP_RSSI_CHAR_VALUE_LENGTH             ( 1 )

/*
 * Hard coded characteristic handles
 *
 * \warning DO NOT CHANGE THESE ATTRIBUTE HANDLE AS THEY REFLECT CERTAIN ATTRIBUTE
 *          CHARACTERISTICS THAT SHOULD BE FOUND AT CERTAIN POSITIONS WITHIN THE
 *          GATT SERVER CREATED. IN DOING SO, CHARACTERISTIC DISCOVERY OPERATIONS
 *          ARE OMITTED, THUS SAVING VALUALE BATTERY ENERGY. */
#define APP_READ_INFO_CHAR_HANDLE              ( 0x0d )
#define APP_WRITE_RSSI_CHAR_HANDLE             ( 0x10 )

/*
 * Define the maximum number of Services running on the device. This will define
 * the array size of the tracked services.
 */
#define APP_MAX_NUM_OF_SERVICES                 ( 10 )

#define MAX_NAME_LEN                            ( BLE_SCAN_RSP_LEN_MAX - 2 )

#define FAST_ADV_SCAN_MIN                       ( 130 ) //! Expressed in ms
#define FAST_ADV_SCAN_MAX                       ( 140 ) //! Expressed in ms

#define POWER_ADV_SCAN_MIN                      ( 320 ) //! Expressed in ms
#define POWER_ADV_SCAN_MAX                      ( 330 ) //! Expressed in ms

/**
 * Min. connection interval in ms.
 *
 * \warning DO NOT CHANGE THAT. OTHERWISE TOF OPERATIONS WILL FAIL (IF ENABLED).
 *
 */
#define CONN_INTERVAL_MIN                       ( 60 )
/**
 * Max. connection interval in ms.
 *
 * \warning DO NOT CHANGE THAT. OTHERWISE TOF OPERATIONS WILL FAIL (IF ENABLED).
 *
 */
#define CONN_INTERVAL_MAX                       ( 70 )
#define CONN_INTERVAL_SUPERVISION_TIMEOUT       ( CONN_INTERVAL_MAX * 8 ) //! Expressed in ms

#define MAX_DISTANCE_TO_RANGE                   ( (float) 300.0 )
#define MIN_DISTANCE_TO_RANGE                   ( (float)   0.0 )

#define DFLT_COARSE_ZONE_DISTANCE_LIMIT         ( (float) 4.0 )
#define DFLT_WARNING_ZONE_DISTANCE_LIMIT        ( (float) 3.0 )
#define DFLT_DANGER_ZONE_DISTANCE_LIMIT         ( (float) 2.0 )

#define DFLT_COARSE_ZONE_RSSI                   ( (int8_t)-80 ) //! Signal strength in dbm --> should be the AVG RSSI expected in line of sight @DFLT_WARNING_ZONE_DISTANCE_LIMIT + 1m
#define DFLT_WARNING_ZONE_RSSI                  ( (int8_t)-65 ) //! Signal strength in dbm --> should be the AVG RSSI expected in line of sight @DFLT_WARNING_ZONE_DISTANCE_LIMIT
#define DFLT_DANGER_ZONE_RSSI                   ( (int8_t)-50 ) //! Signal strength in dbm --> should be the AVG RSSI expected in line of sight @DFLT_DANGER_ZONE_DISTANCE_LIMIT

#define BATTERY_CHECK_INTERVAL                  ( 60 * 1000 ) //! Expressed in ms

/*
 * Min. number of occurrences a device should be scanned in order to
 * be processed for ranging.
 */
#define MIN_SCAN_OCCURENCES                     ( 3 )

#define DFLT_PROXIMITY_SCANNING_DURATION_MS     ( 5000 ) /* The time window performing periodic scans
                                                          * to discover devices and average their RSSI
                                                          * to decide if ranging is needed or not. */

#define PROXIMITY_CONT_OCCURENCES_TO_ALERT      ( 1 )    /* The continuous number of times a device need
                                                          * to be found in WARNING or DANGER zones to
                                                          * trigger alert actions. */

#define AUDIBLE_TONE_HZ                         ( 300 ) //! Expressed in Hz
#define AUDIBLE_TONE_DURATION                   ( 100 ) //! Expressed in ms
#define AUDIBLE_TONES_GAP                       ( 100 ) //! Expressed in ms

#define NUM_OF_WARNING_ALERT_BEEPS              ( 1 )
#define NUM_OF_DANGER_ALERT_BEEPS               ( 3 )
#define MAX_LOCAL_CONN_RSSI_READS               ( 4 )
#define MAX_RF_RANGE_READS                      ( 4 )
#define SERVICES_TO_DISCOVER                    ( 2 )

#if ( SDT_LOGGING_ENABLE == 1 )
        #define MAX_LOG_MESSAGE_SIZE           ( sizeof(log_entry) )
#endif

#define USER_ZONE_X_LOG_SIZE                    ( 512 )


/*********************************************************************
 *
 *      Global variables
 */

static const struct {
        uint16_t min;
        uint16_t max;
} adv_intervals[2] = {
        // "fast connection" interval values
        {
                .min = BLE_ADV_INTERVAL_FROM_MS(FAST_ADV_SCAN_MIN),
                .max = BLE_ADV_INTERVAL_FROM_MS(FAST_ADV_SCAN_MAX),
        },
        // "reduced power" interval values
        {
                .min = BLE_ADV_INTERVAL_FROM_MS(POWER_ADV_SCAN_MIN),
                .max = BLE_ADV_INTERVAL_FROM_MS(POWER_ADV_SCAN_MAX),
        }
};

static uint16_t device_name_len;
static char device_name_buf[MAX_NAME_LEN] = {'S', 'D', 'T', 0 };     /* Must be NULL terminated string,
                                                                      * so 1 byte is for '\0' character
                                                                      * at the end of the string. */

static uint32_t ProximityScanningDurationMS = DFLT_PROXIMITY_SCANNING_DURATION_MS;

__RETAINED_UNINIT static att_uuid_t dte_service_uuid;
__RETAINED_UNINIT static att_uuid_t xchange_info_service_uuid;

__RETAINED_RW static float DangerZoneDistance = DFLT_DANGER_ZONE_DISTANCE_LIMIT;
__RETAINED_RW static float WarningZoneDistance = DFLT_WARNING_ZONE_DISTANCE_LIMIT;
__RETAINED_RW static float CoarseZoneDistance = DFLT_COARSE_ZONE_DISTANCE_LIMIT;

__RETAINED_RW static int8_t DangerZoneRssi = DFLT_DANGER_ZONE_RSSI;
__RETAINED_RW static int8_t WarningZoneRssi = DFLT_WARNING_ZONE_RSSI;
__RETAINED_RW static int8_t CoarseZoneRssi = DFLT_COARSE_ZONE_RSSI;

__RETAINED static uint8_t conn_wd_random_timeout_offset;

__RETAINED_RW volatile static bool sdt_connected            = false;
__RETAINED_RW volatile static bool sdt_connecting           = false;
__RETAINED_RW volatile static bool sdt_connected_as_master  = false;
__RETAINED_RW volatile static bool sdt_advertising          = false;
__RETAINED_RW volatile static bool sdt_scanning             = false;
__RETAINED_RW volatile static bool sdt_ranging              = false;
__RETAINED_RW volatile static bool sdt_charging             = false;

__RETAINED static gap_scan_params_t scan_params[2];

/* Current advertising/scanning settings */
__RETAINED_RW volatile adv_setting_t advertising_setting = ADV_SCAN_INTERVAL_FAST;
__RETAINED_RW volatile adv_setting_t scan_setting = ADV_SCAN_INTERVAL_FAST;

/* Advertising data */
static const gap_adv_ad_struct_t adv_data[] = {
        GAP_ADV_AD_STRUCT_BYTES(GAP_DATA_TYPE_UUID128_LIST_INC,
                0xC6, 0xE9, 0x79, 0x4D, 0xEF, 0x6F, 0x41, 0xE7, 0xA1, 0x7, 0x7F, 0x15, 0x47, 0xD8, 0x43, 0xD6,), //SDT Device UUID Used for distinguishing SDT from other devices
        GAP_ADV_AD_STRUCT_BYTES(GAP_DATA_TYPE_MANUFACTURER_SPEC,
                0,                      //! Device type (To Be Defined), 0 for wearable devices (connect to it)
                0x11, 0x22, 0x33, 0x44, //! Unique ID
                1,                      //! ToF capable indication, 0 --> No TOF, 1 --> TOF
                1,                      //! Connect-to-me indication, 0 --> Do not connect , 1 --> connect
                0xFF),                  //! Reserved for future use
};

/* Device connection parameters */
static const gap_conn_params_t cp = {
        .interval_min  = BLE_CONN_INTERVAL_FROM_MS(CONN_INTERVAL_MIN),
        .interval_max  = BLE_CONN_INTERVAL_FROM_MS(CONN_INTERVAL_MAX),
        .slave_latency = 0,
        .sup_timeout   = BLE_SUPERVISION_TMO_FROM_MS(CONN_INTERVAL_SUPERVISION_TIMEOUT),
};

__RETAINED static uint32_t   default_xtal_trim;
__RETAINED static OS_MUTEX   advertisers_q_mux;
__RETAINED static OS_TIMER   conn_wd_tim_h;
__RETAINED static OS_TIMER   scan_tim_h;
__RETAINED static OS_TASK    alerts_task_h;
__RETAINED static OS_TASK    log_dispatch_task_h;
__RETAINED static OS_TASK    proximity_task_h;      /* Proximity estimation and action task */
__RETAINED static OS_TASK    ranging_task_h;        /* Perform actual ranging task */
__RETAINED static OS_TIMER   proximity_tim_h;       /* Timer to trigger the proximity task
                                                     * every PROXIMITY_DURATION_MS ms. */
__RETAINED OS_TASK sdt_task_h;         /* SDT main FSM task */

__RETAINED static OS_EVENT sdt_connection_event;
__RETAINED static OS_EVENT sdt_rssi_read_event;
__RETAINED static OS_EVENT sdt_get_remote_info_event;

#if ( SDT_LOGGING_ENABLE == 1 )
        #define MAX_NUM_OF_LOG_MSGS     ( 5 )
        __RETAINED static OS_QUEUE log_msgs;
#endif

#if ( SDT_TOF_RANGING_ENABLE == 1 )
        __RETAINED_RW static bool gDevToF = false;
        __RETAINED_RW static bool gIsAnSDTDevice = false;
        __RETAINED_RW static bool gDeviceToBeLogged = false;

        /* Use 4 byte alignment to avoid the unpacking of phase values for calculations */
        __RETAINED static dte_result_t                  i_result __ALIGNED(4);           /* initiator result */
        __RETAINED static dte_result_t                  r_result __ALIGNED(4);           /* responder result */
        __RETAINED float                                cw_distance_offset;              /* in meters */
        __RETAINED static OS_TICK_TIME                  start_tick;
        __RETAINED static OS_TIMER                      dte_timer;                       /* DTE activity timer */

        __RETAINED static config_params_t               cfg_params;
        __RETAINED static gap_dte_params_t              dte_params;

        __RETAINED static OS_EVENT                      rf_ragning_event;

        __RETAINED_RW static peripheral_context_t       ctx = {0};

        __RETAINED_RW volatile static float distance = 0;
#endif

__RETAINED static queue_t advertizers;          /* the rolling list of the devices
                                                 * discovered within the PROXIMITY_DURATION_MS
                                                 * time window */

/* RETAINED array that holds all the Services' handles */
__RETAINED static ble_service_t* custom_services[APP_MAX_NUM_OF_SERVICES];

/* Buffer used to hold the READ_RSSI characteristic value */
__RETAINED_RW int8_t _read_rssi_val[APP_RSSI_CHAR_VALUE_LENGTH] = {-127};

/* Buffer used to hold the WRITE_RSSI value */
__RETAINED_RW int8_t _written_rssi_val[APP_RSSI_CHAR_VALUE_LENGTH] = {-127};

__RETAINED id_info gLocalDevID;
__RETAINED id_info gPeerDevID;

__RETAINED uint8_t services_discovered;

__RETAINED int8_t peer_rssi_val;

/* Variable used to hold device's RSSI value */
__RETAINED int8_t local_rssi_val;

/* Variable used to hold the maximum RSSI value */
__RETAINED int8_t connection_max_rssi_val;

/* Buffer used to hold the actual length of the returned data from read_logs() */
__RETAINED_RW uint32_t _length = USER_ZONE_X_LOG_SIZE;

/* Buffer used to hold the returned value from read_logs() */
__RETAINED_RW uint8_t _read_log_val[USER_ZONE_X_LOG_SIZE] = {0,};

/* Buffer used to hold the zone that is being targeted */
__RETAINED_RW uint8_t _zone = BLE_LIMIT_ZONE;

/* Buffer used to dictates the way that the log pointer is treated after each read_logs() request. Initially the pointer is set to be retained. */
__RETAINED_RW uint8_t _ptr_config = ZONE_PTR_RETAIN;

/* Buffer used to hold the value sent by the connection master to LOG_reset() the specified folder of the logs */
__RETAINED_RW uint8_t _fld_reset = FOLDER_RESET;

/* Default time values */
__RETAINED_RW rtc_time time_cfg = {
        .hour   = RTC_TIME_HOUR,
        .minute = RTC_TIME_MINUTE,
        .sec    = RTC_TIME_SECOND,
        .hsec   = 0, //! Don't care

        .hour_mode = RTC_TIME_DISPLAY_MODE_24H ? RTC_24H_CLK : RTC_12H_CLK
};

/* Default calendar values */
__RETAINED_RW rtc_calendar calendar_cfg = {
        .year   = RTC_CALENDAR_YEAR,
        .month  = RTC_CALENDAR_MONTH,
        .mday   = RTC_CALENDAR_MONTH_DAY,

        /*
         * User is free to map this value as preferred. For instance:
         * 1 --> Monday
         * ...
         * 7 --> Sunday
         */
        .wday   = RTC_CALENDAR_WEEK_DAY
};


/*********************************************************************
 *
 *      Function prototypes
 */

static void set_advertising_interval(adv_setting_t setting);
static bool find_advertiser(const void *data, const void *match_data);
static ble_error_t connect_to_bd_addr_and_trigger_ranging(const bd_address_t *addr, const gap_conn_params_t *conn_params);

static uint8_t find_the_zone_for_RSSI(int8_t rssi);
static uint8_t find_the_zone_for_distance(float distance);

static void handle_evt_gattc_discover_svc(ble_evt_gattc_discover_svc_t *evt);
static void handle_evt_gattc_discover_completed(ble_evt_gattc_discover_completed_t *evt);

#if ( SDT_TOF_RANGING_ENABLE == 1 )
        static void update_dte_status(dte_status_t status);
        static void handle_dte_results_write_req(dte_result_t *result);
        static void auto_xtal_trim(uint8_t dqf, int16_t f_offset, float f_offset_thrs);
        static void dte_timer_cb(OS_TIMER pxTime);
        static void rf_ranging_responder_start(void);
        static void rf_ranging_initiator_start(void);
#endif /* SDT_TOF_RANGING_ENABLE == 1 */

static void release_queue_elelemnt_space(void *data);

/* Overwrite SDK functions declared as weak */
void sys_usb_ext_hook_detach(void)
{
        sdt_charging = false;
        printf("USB unplugged\r\n");
}
void sys_usb_ext_hook_attach(void)
{
        sdt_charging = true;

        ble_gap_adv_stop();
        ble_gap_scan_stop();
        printf("USB Plugged\r\n");
}

static ble_service_t* find_service_by_handle(uint16_t handle)
{
        int i=0;

        for (i = 0; i < APP_MAX_NUM_OF_SERVICES; i++) {
                ble_service_t *svc = custom_services[i];

                if (svc && handle >= svc->start_h && handle <= svc->end_h)
                {
                        return svc;
                }
        }

        return NULL;
}

static void handle_evt_gattc_discover_svc(ble_evt_gattc_discover_svc_t *evt)
{
        if (memcmp(dte_service_uuid.uuid128, evt->uuid.uuid128, ATT_UUID_LENGTH) == 0) {
                gDevToF = true;
                gIsAnSDTDevice = true;
        }

        if (memcmp(xchange_info_service_uuid.uuid128, evt->uuid.uuid128, ATT_UUID_LENGTH) == 0) {
                gIsAnSDTDevice = true;
        }

        services_discovered++;
}

static void handle_evt_gattc_discover_completed(ble_evt_gattc_discover_completed_t *evt)
{
        if (gIsAnSDTDevice && services_discovered == SERVICES_TO_DISCOVER) {
                OS_TASK_NOTIFY(sdt_task_h, RANGE_APP_NOTIFY, OS_NOTIFY_SET_BITS);
        }
}

/******************************************************************************
 * READ INFO Read Request Callback
 *
 */
static att_error_t get_READ_info_value_cb(uint8_t **value, uint16_t *length)
{
        printf("%s\r\n", __func__);

        /* Return the requested data back to the peer device */
        *value  = (uint8_t *)(&gLocalDevID);         // A pointer that points to the returned data
        *length = sizeof(id_info);                   // The size of the returned data, expressed in bytes.

        return (ATT_ERROR_OK);
}

/******************************************************************************
 * WRITE_RSSI Write Request Callback
 *
 */
static att_error_t set_WRITE_RSSI_value_cb(const uint8_t *value, uint16_t length)
{
        /* Clear the current Characteristic Attribute value */
        memset((void *)_written_rssi_val, 0x20, sizeof(_written_rssi_val));

        /* Update the Characteristic Attribute value as requested by the peer device */
        memcpy((void *)_written_rssi_val, (void *)value, length);

        return (ATT_ERROR_OK);
}

/******************************************************************************
 * Retrieve READ_RSSI value from own database
 *
 */
void set_READ_RSSI_value(const uint8_t *value, uint16_t length)
{
        /* Clear the current Characteristic Attribute value */
        memset((void *)_read_rssi_val, 0x20, sizeof(_read_rssi_val));

        /* Update the Characteristic Attribute value as requested by the peer device */
        memcpy((void *)_read_rssi_val, (void *)value, length);
}

#if ( SDT_LOGGING_ENABLE == 1 )
/* ZONE_0_LOG Read Request Callback */
static att_error_t get_ZONE_X_LOG_value_cb(uint8_t **value, uint16_t* length)
{
        /* Retrieve logs of Zone X up to 512 Bytes at a time. */
        int status = LOG_DumpFile(_zone, _read_log_val, USER_ZONE_X_LOG_SIZE, (UINT *)&_length);
        ASSERT_WARNING(status == LOG_STATUS_OK);

        /* length equal to zero signifies end of file */
        if(0 == _length)
        {
                printf("\n\n\rEnd of Zone Log Reached!\n\n\r");
        }

        /* Return the requested data back to the peer device */
        *value  = _read_log_val;                        // A pointer that points to the returned data.
        *length = _length;                              // A pointer that points to the length of the log returned.

        return (ATT_ERROR_OK);
}

/* FLD_RESET Write Request Callback */
static att_error_t set_FLD_RESET_value_cb(const uint8_t *value, uint16_t length)
{
        int status;

        printf("Requested Folder:\t%d\tTruncating...",*value);

        /* Truncate the folder requested */
        status = LOG_TruncateFile(*value);

        if(0 == status)
        {
                printf("SUCCESS.\n\n\r");
        }
        else
        {
                printf("Operation FAILED: Error_Code\t%d.\n\n\r ", status);
        }

        return (ATT_ERROR_OK);
}

/* ZONE Write Request Callback */
static att_error_t set_ZONE_value_cb(const uint8_t *value, uint16_t length)
{

        /* Clear the current Characteristic Attribute value */
        memset((void *)(&_zone), 0x20, sizeof(_zone));

        /* Update the Characteristic Attribute value as requested by the peer device */
        memcpy((void *)(&_zone), (void *)value, length);

        return (ATT_ERROR_OK);
}

/* PTR_CONFIG Write Request Callback */
static att_error_t set_PTR_CONFIG_value_cb(const uint8_t *value, uint16_t length)
{
        int status;

        printf("Requested Index Reset of Folder%d\tResetting...",*value);

        /* Truncate the folder requested */
        status = LOG_ResetDumpIndexer(*value);

        if(0 == status)
        {
                printf("SUCCESS.\n\n\r");
        }
        else
        {
                printf("Operation FAILED: Error_Code\t%d.\n\n\r ", status);
        }

        return (ATT_ERROR_OK);
}
#endif /* SDT_LOGGING_ENABLE == 1 */

#if ( SDT_TOF_RANGING_ENABLE == 1 )
static att_error_t ranging_handle_write_req_cb(const uint8_t *value, uint16_t length)
{
        att_error_t status = ATT_ERROR_ATTRIBUTE_NOT_FOUND;

        if (length == sizeof(dte_result_t)) {
                handle_dte_results_write_req((dte_result_t *) value);
                status = ATT_ERROR_OK;
        } else  {
                printf("INVALID Ranging Data received\r\n");
                status = ATT_ERROR_INVALID_VALUE_LENGTH;
        }
        return (status);
}
#endif /* SDT_TOF_RANGING_ENABLE == 1 */

static void handle_evt_gattc_read_completed(ble_evt_gattc_read_completed_t* evt)
{
        if (evt->status == ATT_ERROR_OK) {
                if (evt->handle == APP_READ_INFO_CHAR_HANDLE)
                {
                        __aeabi_memcpy((uint8_t*)&gPeerDevID, (uint8_t *)evt->value, sizeof(id_info));
                        gDeviceToBeLogged = true;
                        OS_EVENT_SIGNAL(sdt_get_remote_info_event);
                }
        }
}

#if ( SDT_TOF_RANGING_ENABLE == 1 )
static void handle_dte_results_write_req(dte_result_t *result)
{
        if (ctx.dte_status == DTE_STATUS_CALC) {

                __aeabi_memcpy(&i_result, result, sizeof(dte_result_t));

#if ENABLE_CW_DIST_CALC

                uint8_t dqf = 0;
                distance = calc_distance(&i_result, &r_result, &dqf);

                if (dqf == 100) {
                        /* update distance with cw offset */
                        distance = distance - cw_distance_offset;
                        if (distance < 0) {
                                distance = 0;
                        }
                } else {
                        printf("Invalid distance measurement = %.02f\r\n", distance);
                        distance = 0;
                }

                auto_xtal_trim(dqf, i_result.f_offset, 3.0); // i_result ???
#endif /* ENABLE_CW_DIST_CALC */

                update_dte_status(DTE_STATUS_STOP);

                /* wait for next start from initiator */
        } else {
                if (ctx.dte_status != DTE_STATUS_STOP) {
                        ctx.dte_err_code = DTE_ERROR_SYNC_FAILED;
                        /* stop and wait next start from initiator */
                        update_dte_status(DTE_STATUS_STOP);
                }
        }
}

static uint16_t get_xtal_trim(void)
{
        return REG_GETF(CRG_XTAL, CLK_FREQ_TRIM_REG, XTAL32M_TRIM);
}

void set_xtal_trim(uint16_t xtal32m_trim_value)
{
        REG_SETF(CRG_XTAL, CLK_FREQ_TRIM_REG, XTAL32M_TRIM, xtal32m_trim_value);
}

static uint16_t tune_xtal_trim(float f_offset,  float f_offset_thrs)
{
        uint16_t curr_xtal_trim;
        uint16_t new_xtal_trim;
        uint8_t correction_code = 0;


        new_xtal_trim = curr_xtal_trim = get_xtal_trim();

        /* if abs(f_offset) is above threshold -> try trimming */
        if (fabs(f_offset) >= f_offset_thrs) {

                /* - for f_offset > 0 the responder trims down and the initiator up, while
                 * - for f_offset < 0 the responder trims up and the initiator down
                 * Hence we have 4 states (cc = correction_code):
                 *
                 *            |      R         |        I       |
                 *  ----------|----------------+----------------+
                 *  f_off < 0 | cc=00, trim +1 | cc=10, trim -1 |
                 *  f_off > 0 | cc=01, trim -1 | cc=11, trim +1 |
                 *
                 * Note that one entity (I or R) needs to stop earlier with trimming,
                 * e.g. it requires a larger f_offset_thrs.
                 *
                 * Note: USE_PONG = INITIATOR = CENTRAL
                 *
                 */

                correction_code |= cfg_params.role == DTE_ROLE_INITIATOR ? 2 : 0;
                correction_code |= f_offset > 0 ? 1 : 0;

                switch (correction_code)
                {
                case 0:
                case 3:
                        new_xtal_trim += curr_xtal_trim < MAX_XTAL32M_TRIM ? 1 : 0;
                        break;
                case 1:
                case 2:
                        new_xtal_trim -= curr_xtal_trim > 0 ? 1:0;
                        break;
                }
        }

        if (new_xtal_trim != curr_xtal_trim) {
                set_xtal_trim(new_xtal_trim);
        }

        return new_xtal_trim;
}

static void auto_xtal_trim(uint8_t dqf, int16_t f_offset, float f_offset_thrs)
{
        static uint16_t xtrim = 0;

        /* meas_params.use_auto_xtal_trim is derived from NVRAM at FW-start */
        if (cfg_params.flag_auto_xtal32_trim) {
                if (dqf != 0) {

                        uint16_t new_xtrim = tune_xtal_trim(f_offset, f_offset_thrs);
                        if (xtrim != new_xtrim) {
                                xtrim =  new_xtrim;
                        }
                }
        }
}
#endif /* SDT_TOF_RANGING_ENABLE == 1 */

static bool handle_evt_gatts_read_req(ble_evt_gatts_read_req_t* evt)
{
        ble_service_t *svc = find_service_by_handle(evt->handle);

        if (!svc) {
                return false;
        }

        if (svc->read_req) {
                svc->read_req(svc, evt);
        }

        return true;
}

static bool handle_evt_gatts_write_req(ble_evt_gatts_write_req_t* evt)
{
        ble_service_t *svc = find_service_by_handle(evt->handle);

        if (!svc) {
                return false;
        }

        if (svc->write_req) {
                svc->write_req(svc, evt);
        }

        return true;
}

static bool handle_evt_gatts_prepare_write_req(ble_evt_gatts_prepare_write_req_t* evt)
{
        ble_service_t *svc = find_service_by_handle(evt->handle);

        if (!svc) {
                return false;
        }

        if (svc->prepare_write_req) {
                svc->prepare_write_req(svc, evt);
        }

        return true;
}

static bool handle_evt_gatts_event_sent_req(ble_evt_gatts_event_sent_t* evt)
{

        ble_service_t *svc = find_service_by_handle(evt->handle);

        if (!svc) {
                return false;
        }

        if (svc->event_sent) {
                svc->event_sent(svc, evt);
        }

        return true;
}


#if ( SDT_TOF_RANGING_ENABLE == 1 )

void handle_evt_gap_rfmon_report(ble_evt_gap_rfmon_report_t *report)
{
        if (report->status != 0) {
                update_dte_status(DTE_STATUS_STOP);
                distance = 0;
                return; // this is invalid data
        }

        sampling_buffer_t *smpl_buf = (sampling_buffer_t *)report->smpl_buf;

        /* Notify the application that a new report is ready.
         * Since we only use one buffer for now the application
         * can access the buffer through shared memory,
         * otherwise a queue will be needed.
         */
        report_smpl_buf = smpl_buf;

        if (sdt_connected_as_master){
                update_dte_status(DTE_STATUS_CALC);
        } else {
                update_dte_status(DTE_STATUS_START);
                update_dte_status(DTE_STATUS_CALC);
        }

}

static void handle_dte_results_notification(dte_result_t *result)
{
        if (ctx.dte_status == DTE_STATUS_CALC) {

                __aeabi_memcpy(&r_result, result, sizeof(dte_result_t));

#if ENABLE_CW_DIST_CALC
                uint8_t dqf = 0;
                distance = calc_distance(&i_result, &r_result, &dqf);

                if (dqf == 100) {
                        /* update distance with cw offset */
                        distance = distance - cw_distance_offset;
                        if (distance < 0) {
                                distance = 0;
                        }
                } else {
                        printf("Invalid distance measurement dqf=%d\r\n", dqf);
                        distance = 0;
                }

                auto_xtal_trim(dqf, i_result.f_offset, 2.0);
#endif /* ENABLE_CW_DIST_CALC */

                update_dte_status(DTE_STATUS_STOP);

        } else {
                printf("results received with status %d\r\n", ctx.dte_status);
                if (ctx.dte_status != DTE_STATUS_STOP) {
                        ctx.dte_err_code = DTE_ERROR_SYNC_FAILED;
                        /* stop and wait for timeout to start again */
                        update_dte_status(DTE_STATUS_STOP);
                }
        }
}

static void handle_dte_start(void)
{
        OS_TIMER_START(dte_timer, OS_TIMER_FOREVER);

        if (dte_params.is_initiator) {
                start_tick = OS_GET_TICK_COUNT();

                /* reset error code */
                ctx.dte_err_code = DTE_NO_ERROR;

                ble_error_t ble_err = ble_gap_dte_enable(0x0, true);

                if (ble_err != BLE_STATUS_OK) {
                        printf("ble_gap_dte_enable() failed with Status = 0x%02X\r\n", ble_err);
                }

        } else {
                /* reset error code */
                ctx.dte_err_code = DTE_NO_ERROR;
        }
}

static void handle_dte_calc(void)
{
        sampling_buffer_t *smpl_buf = get_rfmon_report();

        if (dte_params.is_initiator) {
                calc_phases(smpl_buf, &i_result);

                /* Free the sampling buffer */
                cmac_iq_sampling_free_buf(smpl_buf);
                ble_gattc_write(0x0,
                                        DTE_DATA_I_RESULT_WRITE,
                                        0,
                                        sizeof(dte_result_t),
                                        (uint8_t *) &i_result);
        } else {
                calc_phases(smpl_buf, &r_result);

                /* Free the sampling buffer */
                cmac_iq_sampling_free_buf(smpl_buf);
                ble_gatts_send_event(0x0,
                                     DTE_DATA_R_RESULT_NOTIFY,
                                     GATT_EVENT_NOTIFICATION,
                                     sizeof(dte_result_t),
                                     (uint8_t *) &r_result);
        }
}

static void handle_dte_calc_error()
{
        sampling_buffer_t *smpl_buf = get_rfmon_report();

        /* Free the sampling buffer */
        cmac_iq_sampling_free_buf(smpl_buf);
}

static void handle_dte_stop(void)
{
        OS_TIMER_STOP(dte_timer, OS_TIMER_FOREVER);
        OS_EVENT_SIGNAL(rf_ragning_event);

        if (dte_params.is_initiator) {
                start_tick = 0;
        } else {
                if (ctx.dte_status == DTE_STATUS_STOP) {
                        return;
                }
        }
}

static void handle_dte_timeout_initiator(void)
{
        if (ctx.dte_status != DTE_STATUS_STOP) {
                printf("DTE timeout at initiator ... status: %d\r\n", ctx.dte_status);
                ctx.dte_err_code = DTE_ERROR_RESULTS_TO;
                update_dte_status(DTE_STATUS_STOP);
        } else {
                if (ctx.dte_err_code != DTE_NO_ERROR) {
                        printf("Initiator stopped with error %d\r\n", ctx.dte_err_code);
                }
        }
}

static void handle_dte_timeout_responder()
{
        printf("DTE timeout at responder ... status:%d\r\n", ctx.dte_status);

        /* if DTE measurement is already stopped a timeout probably occurred but before the
         * handler was called the results were received and DTE measurement was stopped.
         */
        if (ctx.dte_status != DTE_STATUS_STOP) {
                ctx.dte_err_code = DTE_ERROR_RESULTS_TO;
                update_dte_status(DTE_STATUS_STOP);
        }
}

static void update_dte_status(dte_status_t status)
{
        /* peer==NULL when ranging as responder. */
        bool invalid_transition = false;

        if (dte_params.is_initiator) {
                switch (status) {
                case DTE_STATUS_START:

                        if (ctx.dte_status == DTE_STATUS_STOP) {
                                handle_dte_start();

                        } else {
                                invalid_transition = true;
                        }
                        break;
                case DTE_STATUS_CALC:
                        if (ctx.dte_status == DTE_STATUS_START) {
                                handle_dte_calc();
                        } else {
                                handle_dte_calc_error();
                                invalid_transition = true;
                        }
                        break;
                case DTE_STATUS_STOP:
                        handle_dte_stop();
                        break;
                default:
                        invalid_transition = true;
                        break;
                }

                if (!invalid_transition) {
                        ctx.dte_status = status;
                } else {
                        printf("%s(%d): Invalid status transition from %d to %d\r\n", __func__, __LINE__, ctx.dte_status, status);
                }
        } else {
                switch (status) {
                case DTE_STATUS_START:
                        /*
                         * There is a chance that we have failed to receive
                         * a report, so handle the error and allow starting
                         * a new acquisition cycle
                         */
                        if (ctx.dte_status == DTE_STATUS_CALC) {
                                printf("DTE timeout ...\r\n");
                                /* stop timeout and update statistics */
                                ctx.dte_err_code = DTE_ERROR_RESULTS_TO;
                                handle_dte_stop();
                                /* switch state to allow restart */
                                ctx.dte_status = DTE_STATUS_STOP;
                        }
                        if (ctx.dte_status == DTE_STATUS_STOP) {
                               handle_dte_start();

                        } else {
                                invalid_transition = true;
                        }
                        break;
                case DTE_STATUS_CALC:
                        if (ctx.dte_status == DTE_STATUS_START) {
                                handle_dte_calc();
                        } else {
                                handle_dte_calc_error();
                                invalid_transition = true;
                        }
                        break;
                case DTE_STATUS_STOP:
                        handle_dte_stop();
                        break;
                default:
                        invalid_transition = true;
                        break;
                }

                if (!invalid_transition) {
                        ctx.dte_status = status;
                } else {
                        printf("%s(%d): Invalid status transition from %d to %d\r\n", __func__, __LINE__, ctx.dte_status, status);
                }

        }
}

static void handle_evt_gattc_notification(ble_evt_gattc_notification_t *evt)
{
        if (evt->handle == DTE_DATA_R_RESULT_NOTIFY) {
                if (evt->length == sizeof(dte_result_t)) {
                        /* responder results received */
                        handle_dte_results_notification((dte_result_t *) evt->value);
                }
        }
}
#endif /* SDT_TOF_RANGING_ENABLE == 1 */

/**
 * Helper function to get the RSSI of a remote device
 *
 * \return True if the RSSI value was read, false otherwise.
 */
static bool get_peer_connection_rssi(void)
{
        ble_error_t ble_status;

        ble_status = ble_gap_conn_rssi_get(0x0, &local_rssi_val);

        if (ble_status == BLE_STATUS_OK && local_rssi_val != 0 && local_rssi_val != -127) {
                /* Make a write request to the peer for WRITE_RSSI characteristic and pass the local RSSI */
                ble_status = ble_gattc_write(0x0, APP_WRITE_RSSI_CHAR_HANDLE, 0x0, sizeof(local_rssi_val), (uint8_t *)(&local_rssi_val));
                if (ble_status == BLE_STATUS_OK) {
                        return true;
                }

                return false;
        }

        return false;
}

/***********************************************************************
 *
 *  Routine to be called to connect to specific device address
 */
static ble_error_t connect_to_bd_addr_and_trigger_ranging(const bd_address_t *addr, const gap_conn_params_t *conn_params)
{
        ble_error_t status = BLE_ERROR_BUSY;

        printf("Connecting...\r\n");
        if (sdt_advertising){
                ble_gap_adv_stop();
        }

        /* wait for possible incoming connection already started ... */
        OS_DELAY_MS(FAST_ADV_SCAN_MAX);

        if (sdt_connected) {
                sdt_connecting = false;
                return (status);
        }

        ble_gap_scan_stop();

        /* ... no incoming connection established, we are OK */

        status = ble_gap_connect(addr, &cp);

        if (status != BLE_STATUS_OK) {
                sdt_connecting = false;
                printf("%s: Connect command failed. Status=%d\r\n", __func__, status);
                if (status == BLE_ERROR_BUSY) {
                        ble_gap_connect_cancel();
                }

                return (status);
        }

        sdt_connected = false;
        sdt_connecting = true;
        sdt_connected_as_master = false;

        return (status);
}

/* GAP connection event handler */
static void handle_evt_gap_connected(ble_evt_gap_connected_t *evt)
{
        printf("Connected\r\n");

        sdt_connected = true;
        sdt_connecting = false;
        sdt_connected_as_master = false;

        local_rssi_val = -127;
        peer_rssi_val = -127;
        connection_max_rssi_val = -127;

        gDevToF = false;
        gIsAnSDTDevice = false;
        gDeviceToBeLogged = false;

        services_discovered = 0;

        /* SET MTU exchange before discovery */
        ble_gattc_exchange_mtu(evt->conn_idx);

        /* Check if DTE service exists */
        ble_gattc_discover_svc(evt->conn_idx, &dte_service_uuid);

        /* Check if the exchange info service */
        ble_gattc_discover_svc(evt->conn_idx, &xchange_info_service_uuid);

}

/* GAP connection completed event handler */
static void handle_evt_gap_conn_completed(ble_evt_gap_connection_completed_t *evt)
{
        if (evt->status != BLE_STATUS_OK) {
                printf("Master FAILED (%02X)\r\n", evt->status);
                OS_TASK_NOTIFY(proximity_task_h, PROX_CHECK_APP_NOTIFY, OS_NOTIFY_SET_BITS);
        } else {
                sdt_connected_as_master = true;

                OS_EVENT_SIGNAL(sdt_connection_event);
        }
}

/* GAP disconnection event handler */
static void handle_evt_gap_disconnected(ble_evt_gap_disconnected_t *evt)
{
        /* set TX power to MAX db to help establish the next CONNECTION in signal blocking conditions */
        ble_gap_tx_power_set(GAP_AIR_OP_INITIATE, GAP_TX_POWER_MAX);

        printf("Disconnected\r\n");

        sdt_connected = false;
        sdt_connecting = false;
        sdt_connected_as_master = false;

        gDevToF = false;
        gIsAnSDTDevice = false;
        gDeviceToBeLogged = false;

        services_discovered = 0;

        OS_TASK_NOTIFY(sdt_task_h, BLE_START_SCAN_NOTIF, OS_NOTIFY_SET_BITS);

        if (!sdt_advertising && !sdt_charging) {
                set_advertising_interval(ADV_SCAN_INTERVAL_FAST);
                if (ble_gap_adv_start(GAP_CONN_MODE_UNDIRECTED) == BLE_STATUS_OK) {
                        sdt_advertising = true;
                        printf("Adv OK\r\n");
                } else {
                        sdt_advertising = false;
                        printf("Adv FAILED\r\n");
                }
        }
}

/* GAP advertisement completed event handler */
static void handle_evt_gap_adv_completed(ble_evt_gap_adv_completed_t *evt)
{
        if (sdt_connected && !sdt_connected_as_master) {
#if (SDT_TOF_RANGING_ENABLE==1)
                ble_gattc_util_write_ccc(0x0, DTE_DATA_R_RESULT_NOTIFY_ENABLE, GATT_CCC_NOTIFICATIONS);
                rf_ranging_responder_start();
#endif
        }
}

/* GAP pairing event handler */
static void handle_evt_gap_pair_req(ble_evt_gap_pair_req_t *evt)
{
        printf("Pair request\r\n");
        printf("\tConnection index: %d\r\n", evt->conn_idx);
        printf("\tBond: %s\r\n", evt->bond ? "true" : "false");

        ble_gap_pair_reply(evt->conn_idx, true, evt->bond);
}

/* GAP security request event handler */
static void handle_evt_gap_security_request(ble_evt_gap_security_request_t *evt)
{
        ble_error_t status;

        status = ble_gap_pair(evt->conn_idx, evt->bond);

        if (status != BLE_STATUS_OK) {
                printf("%s: failed. Status=%d\r\n", __func__, status);
        }
}

/* Helper function to find the advertiser UUID */
static bool find_advertiser(const void *data, const void *match_data)
{
        adv_dev_t *pAdv_dev = (adv_dev_t *) data;
        uint32_t UID = *((uint32_t *) match_data);

        if (pAdv_dev->id.DevUID == UID) {
                return true;
        } else {
                return false;
        }
}

/* GAP advertising report event handler */
static void handle_evt_gap_adv_report(ble_evt_gap_adv_report_t *adv_report)
{
        adv_dev_t* pAdv_dev = NULL;

        if (memcmp(((sdt_adv_report*)(adv_report->data))->uuid, adv_data->data, 16) == 0) { // check the first 16-bytes UUID for an SDT device

                OS_MUTEX_GET(advertisers_q_mux, OS_MUTEX_FOREVER);
                pAdv_dev = queue_find(&advertizers, find_advertiser, &((sdt_adv_report*)(adv_report->data))->UID);

                if (pAdv_dev != NULL) {
                        /* there is a record and we need to update it */
                        pAdv_dev->num_of_scan_occurences++;
                        pAdv_dev->rssi_sum += adv_report->rssi;
                } else {
                        /* this is a new entry */
                        pAdv_dev = OS_MALLOC(sizeof(adv_dev_t));
                        pAdv_dev->num_of_scan_occurences = 1;
                        pAdv_dev->rssi_sum = adv_report->rssi;
                        pAdv_dev->id.DevType = ((sdt_adv_report*)(adv_report->data))->type;
                        pAdv_dev->id.DevToF = ((sdt_adv_report*)(adv_report->data))->ToF;
                        pAdv_dev->id.DevUID = ((sdt_adv_report*)(adv_report->data))->UID;
                        pAdv_dev->id.Dev_conn_to_me = ((sdt_adv_report*)(adv_report->data))->Conn_to_me;
                        __aeabi_memcpy(&pAdv_dev->bd_addr, &adv_report->address, sizeof(bd_address_t));
                        queue_push_back(&advertizers, pAdv_dev);
                }
                OS_MUTEX_PUT(advertisers_q_mux);
        }
}

/********************************************************************************
 *
 * Routine to set the advertising interval
 *
 * \param[in] setting  Advertising settings reflecting a time frame window
 *
 */
static void set_advertising_interval(adv_setting_t setting)
{
        uint16_t min = adv_intervals[setting].min;
        uint16_t max = adv_intervals[setting].max;

        /* Save current advertising setting */
        advertising_setting = setting;

        ble_gap_adv_intv_set(min, max);
}

/*******************************************************************************
 *
 * Routine to generate the device address
 *
 * At initial run, and if the device address is the SDK default, the routine will generate
 * a random one, set it and save it in the flash memory in the corresponding partition.
 * In doing so, the device will use that value as BD address on boot.
 */
static void update_bd_address(void)
{
        own_address_t addr;

        ble_gap_address_get(&addr);
        if (addr.addr[0] == 0x01 &&
            addr.addr[1] == 0x00 &&
            addr.addr[2] == 0x80 &&
            addr.addr[3] == 0xCA &&
            addr.addr[4] == 0xEA &&
            addr.addr[5] == 0x80)
        {
                uint8_t offset = 0;
                nvms_t nvms;
                /*
                 * in case the current address is equal to the default one, we store a
                 * random address in 0x70000.
                 */
                printf("generating random BD address.\r\n");
                sys_trng_get_bytes(addr.addr, sizeof(addr.addr));
                /* the two MSBs of a random static address have to be 1 */
                addr.addr[5] |= 0xC0;
                addr.addr_type = PUBLIC_STATIC_ADDRESS;
                ble_gap_address_set(&addr, 0x7530); // 5 min

                nvms = ad_nvms_open(NVMS_PARAM_PART);
                if (!nvms) {
                        printf("\r\nCannot open NVMS_PARAM_PART, storing device address failed.\r\n");
                        return;
                }

                /* write random address to NVM and set it to valid */
                ad_nvms_write(nvms, offset, addr.addr, sizeof(addr.addr));
                ad_nvms_write(nvms, 6, &offset, sizeof(offset));
        }
        ble_gap_address_get(&addr);
        /* SDK assumes that a stored address is public so we need to
         * check for random static address and set the correct type.
         */
        if ((addr.addr[BD_ADDR_LEN-1] & 0xC0) == 0xC0) {
                addr.addr_type = PUBLIC_STATIC_ADDRESS;
                ble_gap_address_set(&addr, 0x0);
        }

        printf("Device address: ");
        for (int i= sizeof(addr.addr) - 1;i >=0;i--) {
             printf("%02X",addr.addr[i]);
             if (i != 0) {
                     printf("::");
             } else {
                     printf("\r\n");
             }
        }
        printf("Device address type: %d\r\n", addr.addr_type);
}

static void release_queue_elelemnt_space(void *data)
{
        OS_FREE(data);
}

#if ( SDT_TOF_RANGING_ENABLE == 1 )
/********************************************************************************
 *
 * RF ranging initialization
 */
static void rf_ranging_init(void)
{
        /* Mandatory initialization of the RF ranging system */
        rfmon_init(true, NULL);

        /* Operation configuration parameters */
        cfg_params.role = DTE_ROLE_INITIATOR;
        cfg_params.cw_offset = 0; //! Expressed in cm
        cfg_params.xtal32m_trim_value = get_xtal_trim();
        cfg_params.flag_auto_xtal32_trim = 0;

        /* Set CW distance offset in meters -- global variable for now */
        cw_distance_offset = ((float) cfg_params.cw_offset) / 100;
        printf("CW distance offset %.2f\r\n", cw_distance_offset);

        /* Set default acquisition values */
        dte_params.is_initiator = true;
        dte_params.f_start_mhz = DTE_FREQ_START_MHZ;
        dte_params.f_step_mhz = DTE_FREQ_STEP_MHZ;
        dte_params.nb_atoms = DTE_N_ATOM_MAX;
        dte_params.tx_power = GAP_TX_POWER_0_dBm;

        dte_params.meas_length_us = DTE_ATOM_RX_LENGTH_US;
        dte_params.agc_freeze_lvl = AGC_FREEZE_LEVEL;
        /* If auto freeze is enabled agc_freeze level is ignored */
        dte_params.agc_freeze_auto = true;

        OS_ASSERT(dte_params.nb_atoms * US_TO_BYTES(dte_params.meas_length_us) <= RFMON_IQ_DATA_BUFFER_SIZE);


        /* Create a timer that will be used to timeout DTE measurements which take too long */
        dte_timer = OS_TIMER_CREATE("dte_tim",
                                    OS_MS_2_TICKS(15 /* ToF RF duration */ + (CONN_INTERVAL_MIN * 7) /* the min events + 2*/ + 100 /* the processing time */),
                                    OS_TIMER_FAIL,
                                    NULL,
                                    dte_timer_cb);
}


/******************************************************************************
 *
 *  Routine to be called on connection completion (on master side)
 */
static void rf_ranging_initiator_start(void)
{
        dte_params.is_initiator = true;
        cfg_params.role = DTE_ROLE_INITIATOR;
        ctx.dte_status = DTE_STATUS_STOP;

#if ENABLE_CW_DIST_CALC
        /* Initialize CW distance calculation, buffer address with IQ data is not fixed  */
        cwd_init(dte_params.is_initiator,
                 dte_params.nb_atoms,
                 dte_params.meas_length_us,
                 dte_params.f_step_mhz);
#endif
        /* Send HCI command to set DTE acquisition parameters */
        ble_error_t ble_err = ble_gap_dte_params_set(&dte_params);
        if (ble_err != BLE_STATUS_OK) {
                printf("%s: ble_gap_dte_params_set status = %d", __func__, ble_err);
        }
}

/******************************************************************************
 *
 *  Routine be called on connection completion as slave
 */
static void rf_ranging_responder_start(void)
{
        dte_params.is_initiator = false;
        cfg_params.role = DTE_ROLE_RESPONDER;
        ctx.dte_status = DTE_STATUS_STOP;

#if ENABLE_CW_DIST_CALC
        /* initialize CW distance calculation, buffer address with IQ data is not fixed  */
        cwd_init(dte_params.is_initiator,
                 dte_params.nb_atoms,
                 dte_params.meas_length_us,
                 dte_params.f_step_mhz);
#endif
        /* Send HCI command to set DTE acquisition parameters */
        ble_error_t ble_err = ble_gap_dte_params_set(&dte_params);
        if (ble_err != BLE_STATUS_OK) {
                printf("%s: ble_gap_dte_params_set status = %d", __func__, ble_err);
        }

}
#endif /* SDT_TOF_RANGING_ENABLE == 1 */

#if ( SDT_TOF_RANGING_ENABLE == 1 )
static void dte_timer_cb(OS_TIMER pxTime)
{
        OS_TASK_NOTIFY(sdt_task_h, DTE_TO_NOTIF, OS_NOTIFY_SET_BITS);
}
#endif /* SDT_TOF_RANGING_ENABLE == 1 */

/******************************************************************************
 *
 * Device scan timer callback
 *
 * Timer callback used to trigger periodically scan. The timer is not auto-reload so in callback we can
 * change its period based on the value of the adv_setting variable.
 */
static void scan_tim_cb(OS_TIMER timer)
{
        OS_TASK task = (OS_TASK) OS_TIMER_GET_TIMER_ID(timer);

        OS_TASK_NOTIFY(task, BLE_START_SCAN_NOTIF, OS_NOTIFY_SET_BITS);

        OS_TIMER_CHANGE_PERIOD(scan_tim_h, OS_MS_2_TICKS(scan_setting == ADV_SCAN_INTERVAL_FAST ? 6 * FAST_ADV_SCAN_MIN : 6 * POWER_ADV_SCAN_MIN) , 0);
        OS_TIMER_START(scan_tim_h,0);
}

static void conn_wd_tim_cb(OS_TIMER timer)
{
        if (sdt_connecting && !sdt_connected) {
                OS_EVENT_SIGNAL(sdt_connection_event);
        }
}

/* Proximity check timer callback */
static void proximity_tim_cb(OS_TIMER timer)
{
        OS_TASK_NOTIFY(proximity_task_h, PROX_CHECK_APP_NOTIFY, OS_NOTIFY_SET_BITS);
}

/******************************************************************************
 *
 *  Alert task
 *
 *  Here should add your own alert functionality
 */
static void alerts_task(void *params)
{
        int8_t wdog_id;
        uint8_t num_of_beeps;
        static const timer_config audiable_pwm_timer_cfg = {
                .clk_src = HW_TIMER_CLK_SRC_INT,
                .prescaler = 0,
                .timer = { .direction = HW_TIMER_DIR_UP,
                           .reload_val = 0,
                           .free_run = true },
                .pwm = {
                        .frequency  = configSYSTICK_CLOCK_HZ/AUDIBLE_TONE_HZ,   //! This will make the audible tone
                        .duty_cycle = 1,                                        //! 1 = 50% Duty Cycle
                        .pwm_active_in_sleep = true,
                        .pin  = HW_GPIO_PIN_1,
                        .port = HW_GPIO_PORT_1,
                },
        };

        /* Register ble_multi_link task to be monitored by watchdog */
        wdog_id = sys_watchdog_register(false);

        for (;;) {
                OS_BASE_TYPE ret;
                uint32_t notif;

                /* Notify watchdog on each loop */
                sys_watchdog_notify(wdog_id);

                /* Suspend watchdog while blocking on OS_TASK_NOTIFY_WAIT() */
                sys_watchdog_suspend(wdog_id);

                /*
                 * Wait on any of the notification bits, then clear them all
                 */
                ret = OS_TASK_NOTIFY_WAIT(0, OS_TASK_NOTIFY_ALL_BITS, &notif, OS_TASK_NOTIFY_FOREVER);
                OS_ASSERT(ret == OS_TASK_NOTIFY_SUCCESS);

                /* Resume watchdog */
                sys_watchdog_notify_and_resume(wdog_id);

                num_of_beeps = 0;

                /* Notified from BLE Manager? */
                if (notif & WARNING_ALERT_NOTIFY) {
                        num_of_beeps = NUM_OF_WARNING_ALERT_BEEPS;
                }

                if (notif & DANGER_ALERT_NOTIFY) {
                        num_of_beeps = NUM_OF_DANGER_ALERT_BEEPS;
                }

                printf("\t%dx beep...\r\n", num_of_beeps);

                if (num_of_beeps > 0) {
                        for (int i=0; i<num_of_beeps; i++) {
                                /* Notify watchdog on each loop */
                                sys_watchdog_notify(wdog_id);

                                /* Suspend watchdog while blocking on OS_TASK_NOTIFY_WAIT() */
                                sys_watchdog_suspend(wdog_id);

                                hw_timer_init(HW_TIMER, &audiable_pwm_timer_cfg);
                                hw_gpio_set_pin_function(HW_GPIO_PORT_1, HW_GPIO_PIN_1, HW_GPIO_MODE_OUTPUT, HW_GPIO_FUNC_TIM_PWM);
                                hw_timer_enable(HW_TIMER);
                                OS_DELAY_MS(AUDIBLE_TONE_DURATION);
                                hw_timer_disable(HW_TIMER);
                                OS_DELAY_MS(AUDIBLE_TONES_GAP);

                                /* Resume watchdog */
                                sys_watchdog_notify_and_resume(wdog_id);
                        }
                }
        }
}

/**********************************************************************************
 *
 * Get the peer's RSSI value and send the local RSSI value to  the peer device.
 */
static void get_connection_rssi(int8_t wdog_id)
{
        uint8_t get_local_conn_rssi_retries;

        /* Set TX power to 0 db for measuring CONNECTION RSSI */
        ble_gap_tx_power_set(GAP_AIR_OP_INITIATE, GAP_TX_POWER_0_dBm);

        for (get_local_conn_rssi_retries = 0;
                     get_local_conn_rssi_retries < MAX_LOCAL_CONN_RSSI_READS && sdt_connected; ) {

                /* Notify watchdog on each loop */
                sys_watchdog_notify(wdog_id);

                if (get_peer_connection_rssi()) {
                        /* Suspend watchdog while blocking on OS_TASK_NOTIFY_WAIT() */
                        sys_watchdog_suspend(wdog_id);
                        OS_EVENT_WAIT(sdt_rssi_read_event, OS_MS_2_TICKS(CONN_INTERVAL_MAX*4));
                        /* Resume watchdog */
                        sys_watchdog_notify_and_resume(wdog_id);

                        get_local_conn_rssi_retries++;
                } else {
                        /* Suspend watchdog while blocking on OS_TASK_NOTIFY_WAIT() */
                        sys_watchdog_suspend(wdog_id);
                        /* Failed to read RSSI, wait for one connection interval */
                        OS_DELAY_MS(CONN_INTERVAL_MAX);
                        /* Resume watchdog */
                        sys_watchdog_notify_and_resume(wdog_id);
                }
        }
}

/**********************************************************************************
 *
 * Routine to be called to get the distance between two devices based on ToF
 */
static float get_tof_distance(dte_role_t role, int8_t wdog_id)
{
        float tmp_distance_sum = 0;
        float tmp_distance_count = 0;
        float avg_distance = MAX_DISTANCE_TO_RANGE;

        if (OS_EVENT_CHECK(rf_ragning_event) == OS_EVENT_SIGNALED) OS_EVENT_WAIT(rf_ragning_event, 0); //to clear already signaled event

        if (role == DTE_ROLE_INITIATOR) {
                rf_ranging_initiator_start();
        }

        /* set TX power to MAX db for CONNECTION during exchange of ToF data to help success of ToF */
        ble_gap_tx_power_set(GAP_AIR_OP_INITIATE, GAP_TX_POWER_0_dBm);

        /* Set TX power for the ToF tone */
        dte_params.tx_power = GAP_TX_POWER_0_dBm;
        cfg_params.flag_auto_xtal32_trim = 0;

        for (int i=0; i<MAX_RF_RANGE_READS && sdt_connected; i++) {
                if (role == DTE_ROLE_INITIATOR) {
                        update_dte_status(DTE_STATUS_START);
                }

                /* Notify watchdog on each loop */
                sys_watchdog_notify(wdog_id);

                /* Suspend watchdog while blocking on OS_TASK_NOTIFY_WAIT() */
                sys_watchdog_suspend(wdog_id);

                OS_BASE_TYPE status = OS_EVENT_WAIT(rf_ragning_event, OS_MS_2_TICKS(15 + (CONN_INTERVAL_MIN * 7) + 100 + 10)); // Wait for signaling 10ms more than the Timeout

                /* Resume watchdog */
                sys_watchdog_notify_and_resume(wdog_id);

                if (status == OS_EVENT_SIGNALED) {
                        printf("\t%s Distance measurement #%d = %.02f , pwr=%d, offset=%.02f, r.AGC=%d, i.AGC=%d, r.RSSI=%d, i.RSSI=%d\r\n", distance==0?"Invalid":"",
                                i, distance, dte_params.tx_power, cw_distance_offset, r_result.AGC_gain, i_result.AGC_gain, r_result.ble_rssi,i_result.ble_rssi);

                        if (distance > MIN_DISTANCE_TO_RANGE && distance < MAX_DISTANCE_TO_RANGE) {
                                tmp_distance_sum += distance;
                                tmp_distance_count++;
                        }
                }
        }

        /* Notify watchdog on each loop */
        sys_watchdog_notify(wdog_id);

        if (tmp_distance_count>0) {
                avg_distance = tmp_distance_sum/tmp_distance_count;
        }
        if (avg_distance == MAX_DISTANCE_TO_RANGE) {
               printf("\r\n\tFAILED DISTANCE MEASUEREMENT\r\n");
               return avg_distance;
        }
        if (role == DTE_ROLE_INITIATOR) {
                printf("\r\n\tINITATOR AVG MEASURED DISTANCE = %.02f\r\n", avg_distance);
        } else {
                printf("\r\n\tRESPONDER AVG MEASURED DISTANCE = %.02f\r\n", avg_distance);
        }

        set_xtal_trim(default_xtal_trim); // Set XTAL trimming back to default

        return avg_distance;
}

/**************************************************************
 *
 * Get the distance zone based on the ToF distance measured
 *
 * \param[in]  The average distance measured after performing
 *             ToF measurements.
 *
 * \return The zone calculated. Should be interpreted via LOG_DISTANCE_ZONE.
 */
static uint8_t find_the_zone_for_distance(float distance)
{
        uint8_t tmp_zone = BLE_LIMIT_ZONE;

#if ( SDT_TOF_RANGING_ENABLE == 1 )
        if (distance <= DangerZoneDistance) {
                tmp_zone = DANGER_ZONE;
        } else {
                if ((distance > DangerZoneDistance) &&
                    (distance <= WarningZoneDistance)) {
                        tmp_zone = WARNING_ZONE;
                } else {

                        if ((distance > WarningZoneDistance) &&
                            (distance <= CoarseZoneDistance)) {
                                tmp_zone = COARSE_ZONE;
                        } else {

                                if ((distance > CoarseZoneDistance)) {
                                        tmp_zone = BLE_LIMIT_ZONE;
                                }
                        }
                }
        }
#endif /* SDT_TOF_RANGING_ENABLE == 1 */
        return tmp_zone;
}

/**************************************************************
 *
 * Get the distance zone based on the average RSSI values
 *
 * \param[in]  The average RSSI value measured by invoking both local
 *             and remote values.
 *
 * \return The zone calculated. Should be interpreted via LOG_DISTANCE_ZONE.
 *
 */
static uint8_t find_the_zone_for_RSSI(int8_t rssi)
{
        uint8_t tmp_zone = BLE_LIMIT_ZONE;

        if (rssi != -127 && rssi != 0) {
                if (rssi >= DangerZoneRssi){
                        tmp_zone = DANGER_ZONE;

                } else {
                        if ((rssi < DangerZoneRssi) && (rssi >= WarningZoneRssi)) {
                                tmp_zone = WARNING_ZONE;
                        } else {
                                if ((rssi < WarningZoneRssi) && (rssi >= CoarseZoneRssi)){
                                         tmp_zone = COARSE_ZONE;
                                } else {
                                        if ((rssi < CoarseZoneRssi)){
                                                 tmp_zone = BLE_LIMIT_ZONE;
                                        }
                                }
                        }
                }
        }
        return tmp_zone;
}

/*
 * Routine to be called to trigger the alert task
 *
 * \param[in] tmp_zone  The current distance zone
 *
 */
static void trigger_alert(uint8_t tmp_zone)
{
        uint32_t alert_notify = 0;

        if (tmp_zone == WARNING_ZONE) {
                alert_notify = WARNING_ALERT_NOTIFY;
                printf("\t! WARNING ZONE !\r\n");
        }

        if (tmp_zone == DANGER_ZONE) {
                alert_notify = DANGER_ALERT_NOTIFY;
                printf("\t*** DANGER ZONE ***\r\n");
        }

        if (alert_notify != 0) {
                OS_TASK_NOTIFY(alerts_task_h, alert_notify, OS_NOTIFY_SET_BITS);
        }
}

/******************************************************************************
 *
 *  Ranging task
 */
static void ranging_task(void *params)
{
        int8_t wdog_id;
        OS_BASE_TYPE os_status;
        log_entry tmp_log;

        /* Alert task */
        os_status = OS_TASK_CREATE("alerts",              /* The text name assigned to the task, for
                                                             debug only; not used by the kernel. */
                                alerts_task,              /* The System Initialization task. */
                                ( void * ) 0,             /* The parameter passed to the task. */
                                4096,                     /* The number of bytes to allocate to the
                                                             stack of the task. */
                                OS_TASK_PRIORITY_NORMAL + 2,  /* The priority assigned to the task. */
                                alerts_task_h);           /* The task handle */
        OS_ASSERT(os_status == OS_TASK_CREATE_SUCCESS);

        /* Register ble_multi_link task to be monitored by watchdog */
        wdog_id = sys_watchdog_register(false);

        for (;;) {
                OS_BASE_TYPE ret;
                uint32_t notif;

                /* Notify watchdog on each loop */
                sys_watchdog_notify(wdog_id);

                /* Suspend watchdog while blocking on OS_TASK_NOTIFY_WAIT() */
                sys_watchdog_suspend(wdog_id);

                /* Wait on any of the notification bits, then clear them all */
                ret = OS_TASK_NOTIFY_WAIT(0, OS_TASK_NOTIFY_ALL_BITS, &notif, OS_TASK_NOTIFY_FOREVER);
                OS_ASSERT(ret == OS_TASK_NOTIFY_SUCCESS);

                /* Resume watchdog */
                sys_watchdog_notify_and_resume(wdog_id);

                /* Notified from BLE Manager? */
                if (notif & PERFORM_RANGING_NOTIFY) {
                        float avg_distance = MAX_DISTANCE_TO_RANGE;

                        if (sdt_connected) {
#if ( SDT_TOF_RANGING_ENABLE == 1 )
                                if (sdt_connected_as_master) {
                                        /* perform the ranging as initiator */
                                        /* perform the ToF Ranging as initiator here */
                                        printf("RANGING: master\r\n");
                                        if (gDevToF == 1) {
                                                avg_distance = get_tof_distance(DTE_ROLE_INITIATOR, wdog_id);
                                        }
                                } else {
                                        /* perform the ranging as responder */
                                        /* perform the ToF Ranging as initiator here */
                                        printf("RANGING: slave\r\n");
                                        if (gDevToF == 1) {
                                                avg_distance = get_tof_distance(DTE_ROLE_RESPONDER, wdog_id);
                                        }
                                }

                                OS_DELAY_MS(CONN_INTERVAL_MAX * 4); // Wait for the ToF data to be exchanged with a retransmission opportunity.
#endif /* SDT_TOF_RANGING_ENABLE == 1 */

                                if (sdt_connected_as_master) {
                                        gDeviceToBeLogged = true;
                                } else {
                                        gDeviceToBeLogged = false;
                                        /* Make a read request to the read remote SDT info */
                                        if (ble_gattc_read(0x0, APP_READ_INFO_CHAR_HANDLE, 0x0) == BLE_STATUS_OK) {
                                                OS_EVENT_WAIT(sdt_get_remote_info_event, OS_MS_2_TICKS(CONN_INTERVAL_MAX*4));
                                        }
                                }

                                get_connection_rssi(wdog_id);

                                /* Notify watchdog on each loop */
                                sys_watchdog_notify(wdog_id);

                                OS_DELAY_MS( CONN_INTERVAL_MAX * 2 );

                                printf("Max connection RSSI = %d\r\n", connection_max_rssi_val);

                                /* Notify watchdog on each loop */
                                sys_watchdog_notify(wdog_id);

                                if (connection_max_rssi_val != -127 || avg_distance != MAX_DISTANCE_TO_RANGE) {
                                        uint8_t tmp_rssi_zone = find_the_zone_for_RSSI(connection_max_rssi_val);
                                        uint8_t tmp_dist_zone = find_the_zone_for_distance(avg_distance);

                                        hw_rtc_get_time_clndr(&time_cfg, &calendar_cfg);

                                        tmp_log.start   = 0x7C;

                                        tmp_log.DevType = gLocalDevID.DevType;
                                        tmp_log.DevUID  = gLocalDevID.DevUID;

                                        tmp_log.hour    = time_cfg.hour;
                                        tmp_log.minute  = time_cfg.minute;
                                        tmp_log.sec     = time_cfg.sec;

                                        tmp_log.Zone = (tmp_rssi_zone < tmp_dist_zone) ? tmp_rssi_zone : tmp_dist_zone;

                                        tmp_log.year    = calendar_cfg.year;
                                        tmp_log.month   = calendar_cfg.month;
                                        tmp_log.mday    = calendar_cfg.mday;

                                        tmp_log.ContactDevType = gPeerDevID.DevType;
                                        tmp_log.ContactDevUID  = gPeerDevID.DevUID;

                                        tmp_log.BatPercent = (SDT_BATTERY_MONITORING_ENABLE == 1) ?
                                                                             read_battery_level() :
                                                                                               100;


#if ( SDT_LOGGING_ENABLE == 1 )
                                        if (gDeviceToBeLogged) {
                                                printf("*-*-*-* Log the entry *-*-*-*\r\n");
                                                OS_QUEUE_PUT(log_msgs, &tmp_log, 0);
                                        }
#endif /* SDT_LOGGING_ENABLE == 1 */

                                        /* Notify watchdog on each loop */
                                        sys_watchdog_notify(wdog_id);

                                        /* trigger alerts only for devices to be logged */
                                        trigger_alert(tmp_log.Zone);
                                }
                                /* done with this, disconnect */
                                if (sdt_connected && sdt_connected_as_master) ble_gap_disconnect(0x0, BLE_HCI_ERROR_REMOTE_USER_TERM_CON);
                        }
                        /* Notify watchdog on each loop */
                        sys_watchdog_notify(wdog_id);
                }
                sdt_ranging = false;
        }
}

/******************************************************************************
 *
 *  Proximity task
 */
static void proximity_sw_fsm_task(void *params)
{
        int8_t wdog_id;

        /* Register ble_multi_link task to be monitored by watchdog */
        wdog_id = sys_watchdog_register(false);

        for (;;) {
                OS_BASE_TYPE ret;
                uint32_t notif;

                /* Notify watchdog on each loop */
                sys_watchdog_notify(wdog_id);

                /* Suspend watchdog while blocking on OS_TASK_NOTIFY_WAIT() */
                sys_watchdog_suspend(wdog_id);

                /* Wait on any of the notification bits, then clear them all */
                ret = OS_TASK_NOTIFY_WAIT(0, OS_TASK_NOTIFY_ALL_BITS, &notif, OS_TASK_NOTIFY_FOREVER);
                OS_ASSERT(ret == OS_TASK_NOTIFY_SUCCESS);

                /* Resume watchdog */
                sys_watchdog_notify_and_resume(wdog_id);

                /* Notified from BLE Manager? */
                if (notif & PROX_CHECK_APP_NOTIFY) {
                        adv_dev_t* pAdv_dev = NULL;
                        volatile int avg_rssi = -127;


                        if (!sdt_connecting && !sdt_connected && !sdt_ranging && !sdt_charging) {
                                OS_MUTEX_GET(advertisers_q_mux, OS_MUTEX_FOREVER);
                                pAdv_dev = queue_pop_front(&advertizers);
                                OS_MUTEX_PUT(advertisers_q_mux);
                                if (pAdv_dev != NULL){
                                        if (pAdv_dev->id.Dev_conn_to_me){
                                                /* Notify watchdog on each loop */
                                                sys_watchdog_notify(wdog_id);

                                                if (pAdv_dev->num_of_scan_occurences >= MIN_SCAN_OCCURENCES){
                                                        uint8_t tmp_zone = BLE_LIMIT_ZONE;

                                                        avg_rssi = pAdv_dev->rssi_sum/pAdv_dev->num_of_scan_occurences;

                                                        if (avg_rssi >= DangerZoneRssi){
                                                                tmp_zone = DANGER_ZONE;
                                                        } else {
                                                                if (avg_rssi < DangerZoneRssi && avg_rssi >= WarningZoneRssi){
                                                                        tmp_zone = WARNING_ZONE;
                                                                } else {
                                                                        if (avg_rssi < WarningZoneRssi && avg_rssi >= CoarseZoneRssi){
                                                                                tmp_zone = COARSE_ZONE;
                                                                        }
                                                                }
                                                        }

                                                        if (tmp_zone != BLE_LIMIT_ZONE) {
                                                                if (!sdt_connected && !sdt_connecting && !sdt_ranging) {
                                                                        /* Notify watchdog on each loop */
                                                                        sys_watchdog_notify(wdog_id);
                                                                        /* Suspend watchdog while blocking */
                                                                        sys_watchdog_suspend(wdog_id);

                                                                        if (connect_to_bd_addr_and_trigger_ranging(&pAdv_dev->bd_addr, &cp) == BLE_STATUS_OK) {
                                                                                OS_ENTER_CRITICAL_SECTION();
                                                                                __aeabi_memcpy(&gPeerDevID, &pAdv_dev->id, sizeof(id_info));
                                                                                OS_LEAVE_CRITICAL_SECTION();
                                                                        }

                                                                        /* Resume watchdog */
                                                                        sys_watchdog_notify_and_resume(wdog_id);
                                                                }
                                                        }

                                                        if (sdt_connecting && !sdt_connected) {
                                                                sys_trng_get_bytes(&conn_wd_random_timeout_offset, 1);
                                                                /* Randomize the connection timeout */
                                                               conn_wd_random_timeout_offset = conn_wd_random_timeout_offset & 0x7F; /* limit the number up to 128 */

                                                                /* Notify watchdog on each loop */
                                                                sys_watchdog_notify(wdog_id);

                                                                /* Suspend watchdog while blocking on OS_TASK_NOTIFY_WAIT() */
                                                                sys_watchdog_suspend(wdog_id);

                                                                /* Set period and start the connection WD timer */
                                                                OS_TIMER_CHANGE_PERIOD(conn_wd_tim_h,
                                                                                       OS_MS_2_TICKS(advertising_setting == ADV_SCAN_INTERVAL_FAST ?
                                                                                               ((2 * FAST_ADV_SCAN_MAX) + conn_wd_random_timeout_offset) :
                                                                                               ((2 * POWER_ADV_SCAN_MAX) + conn_wd_random_timeout_offset)),
                                                                                       0);

                                                                OS_TIMER_RESET(conn_wd_tim_h,0); //start the connection WD timer

                                                                OS_EVENT_WAIT(sdt_connection_event, OS_EVENT_FOREVER);

                                                                /* Resume watchdog */
                                                                sys_watchdog_notify_and_resume(wdog_id);

                                                                OS_TIMER_STOP(conn_wd_tim_h,0); //! Stop (if running) the connection WD timer

                                                                if (sdt_connecting) {
                                                                        printf(" ... cancel connection ...\r\n");

                                                                        OS_ENTER_CRITICAL_SECTION();
                                                                        ble_gap_connect_cancel();
                                                                        OS_LEAVE_CRITICAL_SECTION();

                                                                        sdt_connecting = 0;
                                                                        OS_TASK_NOTIFY(proximity_task_h, PROX_CHECK_APP_NOTIFY, OS_NOTIFY_SET_BITS);

                                                                        /* since we failed, start ADV again to allow other devices looging for this one
                                                                         * to connect by the time we will check the next device ... */
                                                                        if (!sdt_advertising && !sdt_connected && !sdt_connecting && !sdt_ranging && !sdt_charging) {
                                                                                set_advertising_interval(ADV_SCAN_INTERVAL_FAST);
                                                                                if (ble_gap_adv_start(GAP_CONN_MODE_UNDIRECTED) == BLE_STATUS_OK) {
                                                                                        sdt_advertising = true;
                                                                                        printf("Adv OK\r\n");
                                                                                } else {
                                                                                        sdt_advertising = false;
                                                                                        printf("Adv FAILED\r\n");
                                                                                }
                                                                        }

                                                                        /* ... and give it a little time to allow successful connection from other devices */
                                                                        uint8_t random_wait, random_wait_coef;
                                                                        sys_trng_get_bytes(&random_wait, 1);
                                                                        sys_trng_get_bytes(&random_wait_coef, 1);
                                                                        random_wait += portTICK_PERIOD_MS; // to make sure no invalid timeout is set
                                                                        random_wait += FAST_ADV_SCAN_MIN;  // Wait unit the ADV interval
                                                                        random_wait_coef = ((random_wait_coef & 0x3) + 2); //to make sure this is not zero and limit it to 2..5
                                                                        OS_DELAY_MS(random_wait * random_wait_coef); // to help breaking the devices actions synchronization
                                                                }
                                                        }

                                                        /* Notify watchdog on each loop */
                                                        sys_watchdog_notify(wdog_id);
                                                }

                                                OS_ENTER_CRITICAL_SECTION();
                                                release_queue_elelemnt_space(pAdv_dev);
                                                pAdv_dev = NULL;
                                                OS_LEAVE_CRITICAL_SECTION();
                                        } else {
                                                OS_TASK_NOTIFY(proximity_task_h, PROX_CHECK_APP_NOTIFY, OS_NOTIFY_SET_BITS);
                                        }
                                }
                        }

                        if (!sdt_advertising && !sdt_connected && !sdt_connecting && !sdt_ranging && !sdt_charging) {
                                set_advertising_interval(ADV_SCAN_INTERVAL_FAST);
                                if (ble_gap_adv_start(GAP_CONN_MODE_UNDIRECTED) == BLE_STATUS_OK) {
                                        sdt_advertising = true;
                                        printf("Adv OK\r\n");
                                } else {
                                        sdt_advertising = false;
                                        printf("Adv FAILED\r\n");
                                }
                        }
                }
        }
}

#if ( SDT_LOGGING_ENABLE == 1 )
/******************************************************************************
 *
 *  Dispatch task
 *
 *  This task is responsible for performing the actual logging into the flash
 *  memory used. All it does is to wait for message in a OS queue and write
 *  them by invoking the logging mechanism.
 *
 */
static void log_dispatch_task(void *params)
{
        int8_t wdog_id;
        log_entry log;

        /* Register ble_multi_link task to be monitored by watchdog */
        wdog_id = sys_watchdog_register(false);

        for (;;) {
                /* Notify watchdog on each loop */
                sys_watchdog_notify(wdog_id);

                /* Suspend watchdog while blocking on OS_TASK_NOTIFY_WAIT() */
                sys_watchdog_suspend(wdog_id);

                /* Wait on data in the Queue */
                if (OS_QUEUE_GET(log_msgs, &log, OS_QUEUE_FOREVER) == OS_QUEUE_OK) {
                        UINT readbytes;

                        /* Resume watchdog */
                        sys_watchdog_notify_and_resume(wdog_id);

                        int status = LOG_WriteData(log.Zone, &log, sizeof(log), &readbytes);
                        ASSERT_WARNING(LOG_STATUS_OK == status);

                } else {
                        /* Resume watchdog */
                        sys_watchdog_notify_and_resume(wdog_id);
                }
        }
}
#endif /* SDT_LOGGING_ENABLE */

/*
 * Load parameters stored in the flash memory. All parameters loaded
 * reflect STD-related configurations. If a parameter is not written
 * in the flash memory, a random one is generated (when necessary).
 *
 */
static void sdt_load_flash_configuration(void)
{
        /* Open SDT NV-Parameters - area. Name is defined in platform_nvparam.h */
        nvparam_t sdt_parameters = ad_nvparam_open("sdt_platform");

        {
                uint8_t sdt_adv_UUID[16];
                uint8_t empty_adv_UUID[16];
                memset(empty_adv_UUID, 0xFF, 16);
                if (ad_nvparam_read(sdt_parameters, NVPARAM_SDT_PLATFORM_ADV_UUID, 16, (void*)sdt_adv_UUID) == 16) {
                        if (memcmp(sdt_adv_UUID, empty_adv_UUID, 16) !=0 ) {
                                memcpy((uint8_t*)(adv_data->data), sdt_adv_UUID, 16);
                        }
                }
        }

        {
                uint8_t sdt_type_r;
                if (ad_nvparam_read(sdt_parameters, NVPARAM_SDT_PLATFORM_TYPE, 1, (void*)&sdt_type_r) == 1) {
                        if (sdt_type_r != 0xFF) {
                                ((uint8_t*)adv_data->data)[16] = sdt_type_r;
                        }

                        gLocalDevID.DevType = ((uint8_t*)adv_data->data)[16];
                }
        }

        {
                uint32_t sdt_UID_r;
                if (ad_nvparam_read(sdt_parameters, NVPARAM_SDT_PLATFORM_UID, 4, (void*)&sdt_UID_r) == 4) {
                        if (sdt_UID_r == 0xFFFFFFFF) {
                                /* generate a random UID*/
                                sys_trng_get_bytes(&sdt_UID_r, 4);

                                /* write the random UID in the FLASH for the next time to exist already */
                                ad_nvparam_write(sdt_parameters, NVPARAM_SDT_PLATFORM_UID, 4, (void*)&sdt_UID_r);
                        }

                        *((uint32_t*)(&adv_data->data[17])) = sdt_UID_r;

                        gLocalDevID.DevUID = *((uint32_t*)(&adv_data->data[17]));
                }
        }

        {
                uint8_t sdt_ToF_r;
                if (ad_nvparam_read(sdt_parameters, NVPARAM_SDT_PLATFORM_TOF, 1, (void*)&sdt_ToF_r) == 1) {
                        if (sdt_ToF_r != 0xFF) {
                                ((uint8_t*)(adv_data->data))[21] = sdt_ToF_r;
                        }

                        gLocalDevID.DevToF = ((uint8_t*)(adv_data->data))[21];
                }
        }


        {
                uint8_t sdt_conn_to_me_r;
                if (ad_nvparam_read(sdt_parameters, NVPARAM_SDT_PLATFORM_CONNECT_TO_ME, 1, (void*)&sdt_conn_to_me_r) == 1) {
                        if (sdt_conn_to_me_r != 0xFF) {
                                ((uint8_t*)(adv_data->data))[22] = sdt_conn_to_me_r;
                        }

                        gLocalDevID.Dev_conn_to_me = ((uint8_t*)(adv_data->data))[22];
                }
        }

        {
                uint8_t sdt_adv_name[MAX_NAME_LEN];
                if (ad_nvparam_read(sdt_parameters, NVPARAM_SDT_PLATFORM_ADV_NAME, MAX_NAME_LEN, (void*)sdt_adv_name) == MAX_NAME_LEN) {
                        char* ff_ptr = memchr(sdt_adv_name, 0xFF, MAX_NAME_LEN);

                        if (ff_ptr == NULL){
                                /* there is a string in the FLASH with max len.*/
                                sdt_adv_name[MAX_NAME_LEN-1] = '\0'; /* NULL terminate it just in case it was not */
                                memcpy(device_name_buf, sdt_adv_name, MAX_NAME_LEN);
                        } else if  ((uint8_t*)ff_ptr != sdt_adv_name) {
                                /* there is a string in the FLASH.*/
                                *ff_ptr = '\0'; /* NULL terminate it just in case it was not */
                                memcpy(device_name_buf, sdt_adv_name, MAX_NAME_LEN);
                        }
                }

                device_name_len = strlen(device_name_buf);
                ASSERT_WARNING(device_name_len <= MAX_NAME_LEN);
        }

        {
                int8_t sdt_danger_rssi_r;
                if (ad_nvparam_read(sdt_parameters, NVPARAM_SDT_PLATFORM_DANGER_RSSI, 1, (void*)&sdt_danger_rssi_r) == 1) {
                        if ((uint8_t)sdt_danger_rssi_r != 0xFF) {
                                DangerZoneRssi = sdt_danger_rssi_r;
                        }
                }
        }

        {
                int8_t sdt_warning_rssi_r;
                if (ad_nvparam_read(sdt_parameters, NVPARAM_SDT_PLATFORM_WARNING_RSSI, 1, (void*)&sdt_warning_rssi_r) == 1) {
                        if ((uint8_t)sdt_warning_rssi_r != 0xFF) {
                                WarningZoneRssi = sdt_warning_rssi_r;
                        }
                }
        }

        {
                int8_t sdt_coarse_rssi_r;
                if (ad_nvparam_read(sdt_parameters, NVPARAM_SDT_PLATFORM_COARSE_RSSI, 1, (void*)&sdt_coarse_rssi_r) == 1) {
                        if ((uint8_t)sdt_coarse_rssi_r != 0xFF) {
                                CoarseZoneRssi = sdt_coarse_rssi_r;
                        }
                }
        }

        {
                uint16_t sdt_danger_distance_r;
                if (ad_nvparam_read(sdt_parameters, NVPARAM_SDT_PLATFORM_DANGER_DISTANCE, 2, (void*)&sdt_danger_distance_r) == 2) {
                        if (sdt_danger_distance_r != 0xFFFF) {
                                DangerZoneDistance = (float)sdt_danger_distance_r/100;
                        }
                }
        }

        {
                uint16_t sdt_warning_distance_r;
                if (ad_nvparam_read(sdt_parameters, NVPARAM_SDT_PLATFORM_WARNING_DISTANCE, 2, (void*)&sdt_warning_distance_r) == 2) {
                        if (sdt_warning_distance_r != 0xFFFF) {
                                WarningZoneDistance = (float)sdt_warning_distance_r/100;
                        }
                }
        }

        {
                uint16_t sdt_coarse_distance_r;
                if (ad_nvparam_read(sdt_parameters, NVPARAM_SDT_PLATFORM_WARNING_DISTANCE, 2, (void*)&sdt_coarse_distance_r) == 2) {
                        if (sdt_coarse_distance_r != 0xFFFF) {
                                CoarseZoneDistance = (float)sdt_coarse_distance_r/100;
                        }
                }
        }


        {
                uint32_t sdt_proximity_duration_r;
                if (ad_nvparam_read(sdt_parameters, NVPARAM_SDT_PLATFORM_PROX_DUR_MS, 4, (void*)&sdt_proximity_duration_r) == 4) {
                        if (sdt_proximity_duration_r != 0xFFFFFFFF) {
                                ProximityScanningDurationMS = sdt_proximity_duration_r;
                        }
                }
        }


        {
                uint8_t sdt_first_boot_r;
                if (ad_nvparam_read(sdt_parameters, NVPARAM_SDT_PLATFORM_FIRST_BOOT, 1, (void*)&sdt_first_boot_r) == 1) {
                        if (sdt_first_boot_r != 0x00) {
                                printf("This is the first run. Going to hibernation mode. This happens only once until the FLASH is fully erased. \r\n");
                                /* this is the first boot */
                                /* Clear the flag and set the sleep mode to hibernation for shipping/storing */
                                sdt_first_boot_r = 0;
                                ad_nvparam_write(sdt_parameters, NVPARAM_SDT_PLATFORM_FIRST_BOOT, 1, &sdt_first_boot_r);
                                pm_sleep_mode_set(pm_mode_hibernation);
                        }
                }
        }
}

/******************************************************************************
 *
 *  Social distance task used to handle all BLE events
 *
 *  This task is responsible for handling all the BLE events reported by the
 *  BLE manager. In addition, any application specific event should be
 *  handled here via the notification mechanism.
 *
 */
void sdt_ble_events_task(void *params)
{
        OS_BASE_TYPE os_status;
        int8_t wdog_id;
        ble_error_t ble_status;
        gap_adv_ad_struct_t *scan_rsp; /* Scan Response object to be populated with <Complete Local Name> AD type */

        cm_sys_clk_set(sysclk_PLL96);
        /* Register ble_multi_link task to be monitored by watchdog */
        wdog_id = sys_watchdog_register(false);

        default_xtal_trim = get_xtal_trim();

#if ( SDT_LOGGING_ENABLE == 1 )
        extern int8_t idle_task_wdog_id;

        printf("Initialize FAT FS\r\n");
        fflush(stdout);
        sys_watchdog_set_latency(wdog_id, 60);
        sys_watchdog_set_latency(idle_task_wdog_id, 60);

        /* Notify watchdog on each loop */
        sys_watchdog_notify(wdog_id);

        /*
         * Initialize the FAT-FS
         */
        int res = FATFS_Init();
        ASSERT_WARNING(res == 0);

        printf("Initialize Logging\r\n");
        /*
         * Initialize the logging mechanism. Should be called at the very
         * beginning and prior to performing any log/dump operation.
         */
        res = LOG_Init();
        ASSERT_WARNING(res == LOG_STATUS_OK)

        /* Notify watchdog on each loop */
        sys_watchdog_notify(wdog_id);
#endif /* SDT_LOGGING_ENABLE == 1 */

        sdt_load_flash_configuration();

        printf("Enable BLE\r\n");
        /* Enable BLE */
        ble_status = ble_enable();
#if ( SDT_TOF_RANGING_ENABLE == 1 )
        printf("Initialize Ranging\r\n");
        rf_ranging_init();
#endif /* SDT_TOF_RANGING_ENABLE == 1 */

        if (ble_status == BLE_STATUS_OK) {
                /* Set both roles */
                ble_gap_role_set(GAP_PERIPHERAL_ROLE | GAP_CENTRAL_ROLE);
        } else {
                printf("%s: failed. Status=%d\r\n", __func__, ble_status);
                ASSERT_ERROR(0);
        }

        /* Register task to BLE framework to receive BLE event notifications */
        ble_register_app();

        update_bd_address();

        /*
         * Set maximum allowed MTU to increase results throughput.
         * Max data packet payload is 251 bytes.
         */
        ble_gap_mtu_size_set(512);

        /* Set TX power for ADV */
        ble_gap_tx_power_set(GAP_AIR_OP_ADV, GAP_TX_POWER_MAX);

        /* Set TX power for CONNECTION */
        ble_gap_tx_power_set(GAP_AIR_OP_INITIATE, GAP_TX_POWER_MAX);

        /* Define Scan Response object internals dealing with retrieved name */
        scan_rsp = GAP_ADV_AD_STRUCT_DECLARE(GAP_DATA_TYPE_LOCAL_NAME, device_name_len, device_name_buf);

        uint8_t pos = 0;

        const mcs_characteristic_config_t Conn_Info_Exchanger[] = {

                /* Initialized Characteristic Attribute */
                CHARACTERISTIC_DECLARATION(8522C64B-ED9D-4789-A95D-000000000001,
                                           sizeof(id_info),
                                           CHAR_WRITE_NO_RESP_PROP_DIS,
                                           CHAR_WRITE_PROP_DIS,
                                           CHAR_READ_PROP_EN,
                                           CHAR_NOTIF_NONE,
                                           REMOTE_READ_INFO,
                                           get_READ_info_value_cb,
                                           NULL,
                                           NULL),

                CHARACTERISTIC_DECLARATION(8522C64B-ED9D-4789-A95D-000000000002,
                                           APP_RSSI_CHAR_VALUE_LENGTH,
                                           CHAR_WRITE_NO_RESP_PROP_DIS,
                                           CHAR_WRITE_PROP_EN,
                                           CHAR_READ_PROP_DIS,
                                           CHAR_NOTIF_NONE,
                                           REMOTE_WRITE_RSSI,
                                           NULL,
                                           set_WRITE_RSSI_value_cb,
                                           NULL),
        };
        // ***************** Register the Bluetooth Service in Dialog BLE framework *****************
        custom_services[pos++] = SERVICE_DECLARATION(Conn_Info_Exchanger, 8522C64B-ED9D-4789-A95D-0F56E9246D4E);
        ble_uuid_from_string("8522C64B-ED9D-4789-A95D-0F56E9246D4E", &xchange_info_service_uuid);

        /*****************************************************************************************************
         *
         *   WARING: Any BLE service, either custom or SIG one, must be declared from this point onwards!!!
         *
         ****************************************************************************************************/

#if ( SDT_TOF_RANGING_ENABLE == 1 )
        const mcs_characteristic_config_t Ranging[] = {

                /* Initialized Characteristic Attribute */
                CHARACTERISTIC_DECLARATION(4F120CA4-85D3-11E8-ABC0-000000000001,
                                           sizeof(dte_result_t),
                                           CHAR_WRITE_NO_RESP_PROP_EN,
                                           CHAR_WRITE_PROP_EN,
                                           CHAR_READ_PROP_DIS,
                                           CHAR_NOTIF_NONE,
                                           DATA_I_RESULT,
                                           NULL,
                                           ranging_handle_write_req_cb,
                                           NULL),

                CHARACTERISTIC_DECLARATION(4F120CA4-85D3-11E8-ABC0-000000000002,
                                           sizeof(dte_result_t),
                                           CHAR_WRITE_NO_RESP_PROP_DIS,
                                           CHAR_WRITE_PROP_DIS,
                                           CHAR_READ_PROP_DIS,
                                           CHAR_NOTIF_NOTIF_EN,
                                           DATA_R_RESULT,
                                           NULL,
                                           NULL,
                                           NULL),
        };
        // ***************** Register the Bluetooth Service in Dialog BLE framework *****************
        custom_services[pos++] = SERVICE_DECLARATION(Ranging, 4F120CA4-85D3-11E8-ABC0-FA7AE01BBEBC);
        ble_uuid_from_string("4F120CA4-85D3-11E8-ABC0-FA7AE01BBEBC", &dte_service_uuid);
#endif

#if ( SDT_LOGGING_ENABLE == 1 )
        const mcs_characteristic_config_t Log_Dump[] = {

                CHARACTERISTIC_DECLARATION(3112fb86-86c7-11ea-bc55-000000000003, USER_ZONE_X_LOG_SIZE,
                        CHAR_WRITE_NO_RESP_PROP_DIS, CHAR_WRITE_PROP_DIS, CHAR_READ_PROP_EN, CHAR_NOTIF_NONE,
                        ZONE_X_LOG, get_ZONE_X_LOG_value_cb, NULL, NULL),

                CHARACTERISTIC_DECLARATION(3112fb86-86c7-11ea-bc55-000000000004, sizeof(_zone),
                        CHAR_WRITE_NO_RESP_PROP_DIS, CHAR_WRITE_PROP_EN, CHAR_READ_PROP_DIS, CHAR_NOTIF_NONE,
                        ZONE, NULL, set_ZONE_value_cb, NULL),

                CHARACTERISTIC_DECLARATION(3112fb86-86c7-11ea-bc55-000000000005, sizeof(_ptr_config),
                        CHAR_WRITE_NO_RESP_PROP_DIS, CHAR_WRITE_PROP_EN, CHAR_READ_PROP_DIS, CHAR_NOTIF_NONE,
                        PTR_CONFIG, NULL, set_PTR_CONFIG_value_cb, NULL),

                CHARACTERISTIC_DECLARATION(3112fb86-86c7-11ea-bc55-000000000006, sizeof(_fld_reset),
                        CHAR_WRITE_NO_RESP_PROP_DIS, CHAR_WRITE_PROP_EN, CHAR_READ_PROP_DIS, CHAR_NOTIF_NONE,
                        FLD_RESET, NULL, set_FLD_RESET_value_cb, NULL),

        };
        // ***************** Register the Bluetooth Service in Dialog BLE framework *****************
        custom_services[pos++] = SERVICE_DECLARATION(Log_Dump, 3112fb86-86c7-11ea-bc55-0242ac130003);
#endif /* SDT_LOGGING_ENABLE == 1 */

        /* Add CTS service */
        cts_init(NULL, &cts_callbacks);

#if ( SDT_BATTERY_MONITORING_ENABLE == 1 )
        /* Add Battery Service */
        bas = bas_init(NULL, NULL);
#endif /* SDT_BATTERY_MONITORING_ENABLE == 1 */

        queue_init(&advertizers);

        sdt_task_h = OS_GET_CURRENT_TASK();

        OS_EVENT_CREATE(sdt_connection_event);
        OS_EVENT_CREATE(sdt_rssi_read_event);
        OS_EVENT_CREATE(sdt_get_remote_info_event);
#if ( SDT_TOF_RANGING_ENABLE == 1 )
        OS_EVENT_CREATE(rf_ragning_event);
#endif /* SDT_TOF_RANGING_ENABLE == 1 */

        OS_MUTEX_CREATE(advertisers_q_mux);

#if ( SDT_LOGGING_ENABLE == 1 )
        OS_QUEUE_CREATE(log_msgs, sizeof(log_entry), MAX_NUM_OF_LOG_MSGS);
#endif /* SDT_LOGGING_ENABLE == 1 */

        /* Set the scan params to follow the ADV params */
        scan_params[ADV_SCAN_INTERVAL_FAST].window = adv_intervals[ADV_SCAN_INTERVAL_FAST].min/3;
        scan_params[ADV_SCAN_INTERVAL_FAST].interval = (adv_intervals[ADV_SCAN_INTERVAL_FAST].max * 3)/2;

        scan_params[ADV_SCAN_INTERVAL_POWER].window = adv_intervals[ADV_SCAN_INTERVAL_POWER].min/3;
        scan_params[ADV_SCAN_INTERVAL_POWER].interval = (adv_intervals[ADV_SCAN_INTERVAL_POWER].max * 3)/2;

        /* Connection establishment watchdog timer */
        conn_wd_tim_h = OS_TIMER_CREATE("conn WD",
                                   OS_MS_2_TICKS(FAST_ADV_SCAN_MAX  + conn_wd_random_timeout_offset),
                                   pdFALSE,
                                   (void *) OS_GET_CURRENT_TASK(),
                                   conn_wd_tim_cb);
        OS_TIMER_START(conn_wd_tim_h, 0);

        /* Trigger the periodic scan */
        scan_tim_h = OS_TIMER_CREATE("scan",
                                   OS_MS_2_TICKS(scan_setting == ADV_SCAN_INTERVAL_FAST ?
                                                                 (6 * FAST_ADV_SCAN_MAX) :
                                                                 (6 * POWER_ADV_SCAN_MAX)),
                                   OS_TIMER_FAIL,
                                   (void *) OS_GET_CURRENT_TASK(),
                                   scan_tim_cb);
        OS_TIMER_START(scan_tim_h, 0);

#if ( SDT_BATTERY_MONITORING_ENABLE == 1 )
        /* Create timer for battery monitoring */
        bas_tim = OS_TIMER_CREATE("bas",
                                  OS_MS_2_TICKS(BATTERY_CHECK_INTERVAL),
                                  pdTRUE,
                                  (void *) OS_GET_CURRENT_TASK(),
                                  bas_tim_cb);
        OS_TIMER_START(bas_tim, OS_TIMER_FOREVER);

        /* Update battery level exposed in BAS */
        bas_update(sdt_charging);
#endif /* SDT_BATTERY_MONITORING_ENABLE == 1 */

        /* Periodic proximity check timer */
        proximity_tim_h = OS_TIMER_CREATE("proximity",
                                   OS_MS_2_TICKS(ProximityScanningDurationMS),
                                   pdTRUE,
                                   (void *) OS_GET_CURRENT_TASK(),
                                   proximity_tim_cb);
        OS_TIMER_START(proximity_tim_h, 0);

#if ( SDT_LOGGING_ENABLE == 1 )
        /* LOG Msg Dispatching */
        os_status = OS_TASK_CREATE("Log Dispatch",        /* The text name assigned to the task, for
                                                             debug only; not used by the kernel. */
                                log_dispatch_task,        /* The System Initialization task. */
                                ( void * ) 0,             /* The parameter passed to the task. */
                                1024,                     /* The number of bytes to allocate to the
                                                             stack of the task. */
                                OS_TASK_PRIORITY_NORMAL,  /* The priority assigned to the task. */
                                log_dispatch_task_h);     /* The task handle */
        OS_ASSERT(os_status == OS_TASK_CREATE_SUCCESS);
#endif /* SDT_LOGGING_ENABLE == 1 */

        /* Proximity check task (main SW FSM) */
        os_status = OS_TASK_CREATE("proximity",           /* The text name assigned to the task, for
                                                             debug only; not used by the kernel. */
                                proximity_sw_fsm_task,    /* The System Initialization task. */
                                ( void * ) 0,             /* The parameter passed to the task. */
                                4096,                     /* The number of bytes to allocate to the
                                                             stack of the task. */
                                OS_TASK_PRIORITY_NORMAL,  /* The priority assigned to the task. */
                                proximity_task_h);        /* The task handle */
        OS_ASSERT(os_status == OS_TASK_CREATE_SUCCESS);

        /* Ranging task */
        os_status = OS_TASK_CREATE("ranging",                   /* The text name assigned to the task, for
                                                                   debug only; not used by the kernel. */
                                ranging_task,                   /* The System Initialization task. */
                                ( void * ) 0,                   /* The parameter passed to the task. */
                                0x2000,                         /* The number of bytes to allocate to the
                                                                   stack of the task. */
                                SDT_RANGING_TASK_PRIORITY,      /* The priority assigned to the task. */
                                ranging_task_h);                /* The task handle */
        OS_ASSERT(os_status == OS_TASK_CREATE_SUCCESS);

        set_advertising_interval(ADV_SCAN_INTERVAL_FAST);

        /* Set advertising data and start advertising */
        ble_gap_adv_ad_struct_set(ARRAY_LENGTH(adv_data), adv_data, 1 , scan_rsp);

        if (!sdt_charging) {
                if (ble_gap_adv_start(GAP_CONN_MODE_UNDIRECTED) == BLE_STATUS_OK) {
                        sdt_advertising = true;
                        printf("Adv OK\r\n");
                } else {
                        sdt_advertising = false;
                        printf("Adv FAILED\r\n");
                }
        }

        cm_sys_clk_set(sysclk_XTAL32M);

        for (;;) {
                static OS_BASE_TYPE ret;
                static uint32_t notif;

                /* Notify watchdog on each loop */
                sys_watchdog_notify(wdog_id);

                /* Suspend watchdog while blocking on OS_TASK_NOTIFY_WAIT() */
                sys_watchdog_suspend(wdog_id);

                /* Wait on any of the notification bits, then clear them all */
                ret = OS_TASK_NOTIFY_WAIT(0, OS_TASK_NOTIFY_ALL_BITS, &notif, OS_TASK_NOTIFY_FOREVER);
                OS_ASSERT(ret == OS_TASK_NOTIFY_SUCCESS);

                /* Resume watchdog */
                sys_watchdog_notify_and_resume(wdog_id);

                /* Notified from BLE Manager? */
                if (notif & BLE_APP_NOTIFY_MASK) {
                        ble_evt_hdr_t *hdr=NULL;

                        hdr = ble_get_event(false);
                        if (!hdr) {
                                goto no_event;
                        }

                        switch (hdr->evt_code) {
#if ( SDT_TOF_RANGING_ENABLE == 1 )
                        /* =============================== HANDLE GAP RFMON SERVER EVENTS ================================= */
                        case BLE_EVT_GAP_RFMON_REPORT:
                                handle_evt_gap_rfmon_report((ble_evt_gap_rfmon_report_t *) hdr);
                                break;
#endif /* SDT_LOGGING_ENABLE == 1 */
                        /* =================================== HANDLE GAP SERVER EVENTS =================================== */

                        case BLE_EVT_GAP_CONNECTED:
                                handle_evt_gap_connected((ble_evt_gap_connected_t *) hdr);
                                break;

                        case BLE_EVT_GAP_CONNECTION_COMPLETED:
                                handle_evt_gap_conn_completed((ble_evt_gap_connection_completed_t *) hdr);
                                break;

                        case BLE_EVT_GAP_DISCONNECTED:
                                handle_evt_gap_disconnected((ble_evt_gap_disconnected_t *) hdr);
                                break;

                        case BLE_EVT_GAP_ADV_COMPLETED:
                                handle_evt_gap_adv_completed((ble_evt_gap_adv_completed_t *) hdr);
                                sdt_advertising = false;
                                break;

                        case BLE_EVT_GAP_PAIR_REQ:
                                handle_evt_gap_pair_req((ble_evt_gap_pair_req_t *) hdr);
                                break;

                        case BLE_EVT_GAP_SECURITY_REQUEST:
                                handle_evt_gap_security_request((ble_evt_gap_security_request_t *) hdr);
                                break;

                        case BLE_EVT_GAP_ADV_REPORT:
                                handle_evt_gap_adv_report((ble_evt_gap_adv_report_t *) hdr); /* scan results come here */
                                break;

                        case BLE_EVT_GAP_SCAN_COMPLETED:
                                sdt_scanning = false;
                                break;

                        /* =================================== HANDLE GATT SERVER EVENTS =================================== */

                        case BLE_EVT_GATTC_READ_COMPLETED:
                                handle_evt_gattc_read_completed((ble_evt_gattc_read_completed_t *) hdr);
                                break;

                        case BLE_EVT_GATTS_READ_REQ:
                                if (!handle_evt_gatts_read_req((ble_evt_gatts_read_req_t *) hdr))
                                {
                                        ble_service_handle_event(hdr);
                                }
                                break;

                        case BLE_EVT_GATTS_WRITE_REQ:
                                if (handle_evt_gatts_write_req((ble_evt_gatts_write_req_t *) hdr))
                                {
                                        if (((ble_evt_gatts_write_req_t *) hdr)->handle == APP_WRITE_RSSI_CHAR_HANDLE)
                                        {
                                                int8_t tmp_connection_max_rssi_val;

                                                /* If this point is reached then the device is the slave/server, so request the RSSI as slave */
                                                ble_gap_conn_rssi_get(((ble_evt_gatts_write_req_t *) hdr)->conn_idx, &local_rssi_val);

                                                /* Write the value that was returned to READ_RSSI characteristic */
                                                set_READ_RSSI_value((uint8_t *)(&local_rssi_val), sizeof(local_rssi_val));

                                                /* Take the rssi sent by the master to be written to WRITE_RSSI */
                                                peer_rssi_val = *((int8_t*)(((ble_evt_gatts_write_req_t *) hdr)->value));

                                                /* Compare the two values and keep the maximum of the two */
                                                tmp_connection_max_rssi_val = (local_rssi_val > peer_rssi_val) ? local_rssi_val : peer_rssi_val;

                                                if (tmp_connection_max_rssi_val > connection_max_rssi_val) {
                                                        connection_max_rssi_val = tmp_connection_max_rssi_val;
                                                }

                                                OS_EVENT_SIGNAL(sdt_rssi_read_event);
                                        }
                                } else {
                                        ble_service_handle_event(hdr);
                                }
                                break;

                        case BLE_EVT_GATTS_PREPARE_WRITE_REQ:
                                if (!handle_evt_gatts_prepare_write_req((ble_evt_gatts_prepare_write_req_t *) hdr)) {
                                        ble_service_handle_event(hdr);
                                }
                                break;

                        case BLE_EVT_GATTS_EVENT_SENT:
                                if (!handle_evt_gatts_event_sent_req((ble_evt_gatts_event_sent_t *) hdr)) {
                                        ble_service_handle_event(hdr);
                                }
                                break;

                        /* =================================== HANDLE GATT CLIENT EVENTS =================================== */
#if ( SDT_TOF_RANGING_ENABLE == 1 )
                        case BLE_EVT_GATTC_NOTIFICATION:
                                handle_evt_gattc_notification((ble_evt_gattc_notification_t *) hdr);
                                break;
#endif /* SDT_TOF_RANGING_ENABLE == 1 */

                        /* ================================== DISCOVER SERVICES/CHARACTERISTICS ============================ */
                        case BLE_EVT_GATTC_DISCOVER_SVC:
                                handle_evt_gattc_discover_svc((ble_evt_gattc_discover_svc_t *) hdr);
                                break;
                        case BLE_EVT_GATTC_DISCOVER_COMPLETED:
                                handle_evt_gattc_discover_completed((ble_evt_gattc_discover_completed_t *) hdr);
                                break;

                        default:
                                ble_service_handle_event(hdr);
                                ble_handle_event_default(hdr);
                                break;
                        }

                        OS_FREE(hdr);

no_event:
                        /* Notify again if there are more events to process in queue */
                        if (ble_has_event()) {
                                OS_TASK_NOTIFY(OS_GET_CURRENT_TASK(), BLE_APP_NOTIFY_MASK,
                                                                               OS_NOTIFY_SET_BITS);
                                continue;
                        }
                }

                /* Timer triggered scan */
                if ((notif & BLE_START_SCAN_NOTIF) && !(sdt_connecting) && !(sdt_connected) && !(sdt_charging)) {
                        ble_gap_scan_params_set(&scan_params[scan_setting]);
                        sdt_scanning = true;
                        ble_gap_scan_start(GAP_SCAN_PASSIVE, GAP_SCAN_GEN_DISC_MODE, scan_params[scan_setting].interval, scan_params[scan_setting].window, false, false);
                }

                if (notif & RANGE_APP_NOTIFY){
                        /*
                         * The device is found with average RSSI over PROXIMITY_SCANNING_DURATION_MS/1000 seconds in WARNING or in DANGER Zones.
                         * Ranging is needed to check more accurately the distance
                         * and take action if this is the PROXIMITY_CONT_OCCURENCES_TO_ALERT
                         * continuous time found in WARNING or DANGER ZONE
                         */
                        if (sdt_connected) {
                                sdt_ranging = true;

                                ble_gap_scan_stop();
                                ble_gap_adv_stop();

                                OS_TASK_NOTIFY(ranging_task_h, PERFORM_RANGING_NOTIFY, OS_NOTIFY_SET_BITS);
                        }
                }

#if ( SDT_TOF_RANGING_ENABLE == 1 )
                if (notif & DTE_TO_NOTIF) {
                        printf("DTE Timeout\r\n");
                        distance = 0;
                        if (dte_params.is_initiator){
                                handle_dte_timeout_initiator();
                        } else {
                                handle_dte_timeout_responder();
                        }
                        OS_EVENT_SIGNAL(rf_ragning_event);
                }
#endif /* SDT_TOF_RANGING_ENABLE == 1 */

#if ( SDT_BATTERY_MONITORING_ENABLE == 1 )
                /* Notified from battery timer? */
                if (notif & BAS_TMO_NOTIF) {
                        /* update battery level */
                        bas_update(sdt_charging);
                }
#endif /* SDT_BATTERY_MONITORING_ENABLE == 1 */

                /* Notify again if there are more events to process in queue */
                if (ble_has_event()) {
                        OS_TASK_NOTIFY(OS_GET_CURRENT_TASK(), BLE_APP_NOTIFY_MASK,
                                                                       OS_NOTIFY_SET_BITS);
                }
        }
}
