/**
 ****************************************************************************************
 *
 * @file pcb_hw_id.c
 *
 * @brief Custom mechanism for ID the PCB revision through the P0_28-P0_31
 *        P0_28 is the LSB
 *
 *                        +-------+-------+-------+-------+
 *        eg.             | P0_31 | P0_30 | P0_29 | P0_28 |
 *                        +-------+-------+-------+-------+
 *  pcb ver. 5 --> 0x5  = |   0   |   1   |   0   |   1   |
 *                        +-------+-------+-------+-------+
 *
 * Copyright (c) 2020 Dialog Semiconductor. All rights reserved.
 *
 * This software ("Software") is owned by Dialog Semiconductor. By using this Software
 * you agree that Dialog Semiconductor retains all intellectual property and proprietary
 * rights in and to this Software and any use, reproduction, disclosure or distribution
 * of the Software without express written permission or a license agreement from Dialog
 * Semiconductor is strictly prohibited. This Software is solely for use on or in
 * conjunction with Dialog Semiconductor products.
 *
 * EXCEPT AS OTHERWISE PROVIDED IN A LICENSE AGREEMENT BETWEEN THE PARTIES OR AS
 * REQUIRED BY LAW, THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. EXCEPT AS OTHERWISE PROVIDED
 * IN A LICENSE AGREEMENT BETWEEN THE PARTIES OR BY LAW, IN NO EVENT SHALL DIALOG
 * SEMICONDUCTOR BE LIABLE FOR ANY DIRECT, SPECIAL, INDIRECT, INCIDENTAL, OR
 * CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR
 * PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THE SOFTWARE.
 *
 ****************************************************************************************
 */

#include "sdk_defs.h"
#include "hw_gpio.h"
#include "hw_sys.h"

/*********************************************************************
 *
 *      Defines
 */

#define HW_ID_PORT              HW_GPIO_PORT_0

#define HW_ID_BIT_0_PIN         HW_GPIO_PIN_28
#define HW_ID_BIT_1_PIN         HW_GPIO_PIN_29
#define HW_ID_BIT_2_PIN         HW_GPIO_PIN_30
#define HW_ID_BIT_3_PIN         HW_GPIO_PIN_31

/*********************************************************************
 *
 *      Retained code/variables
 */
__RETAINED_RW uint8_t pcb_version = 0xFF;       /* 0xFF is an invalid value. The SW should call the
                                                 * get_pcb_version() to get the PCB HW*/

/*
 * \brief Get PCB version
 *
 * Read the status of the GPIOs P0_28 to P0_31 to get the version of the PCB of the device.
 * This can be used in the future versions of the HW/SW to configure the correct pins for
 * each interface.
 */
void get_pcb_version(void)
{
        pcb_version = 0;

        hw_sys_pd_com_enable();

        /* BIT 0 */
        hw_gpio_set_pin_function(HW_ID_PORT, HW_ID_BIT_0_PIN, HW_GPIO_MODE_INPUT, HW_GPIO_FUNC_GPIO);
        hw_gpio_pad_latch_enable(HW_ID_PORT, HW_ID_BIT_0_PIN);
        if (hw_gpio_get_pin_status(HW_ID_PORT, HW_ID_BIT_0_PIN)) pcb_version = pcb_version | (1<<0);
        hw_gpio_pad_latch_disable(HW_ID_PORT, HW_ID_BIT_0_PIN);

        /* BIT 1 */
        hw_gpio_set_pin_function(HW_ID_PORT, HW_ID_BIT_1_PIN, HW_GPIO_MODE_INPUT, HW_GPIO_FUNC_GPIO);
        hw_gpio_pad_latch_enable(HW_ID_PORT, HW_ID_BIT_1_PIN);
        if (hw_gpio_get_pin_status(HW_ID_PORT, HW_ID_BIT_1_PIN)) pcb_version = pcb_version | (1<<1);
        hw_gpio_pad_latch_disable(HW_ID_PORT, HW_ID_BIT_1_PIN);

        /* BIT 2 */
        hw_gpio_set_pin_function(HW_ID_PORT, HW_ID_BIT_2_PIN, HW_GPIO_MODE_INPUT, HW_GPIO_FUNC_GPIO);
        hw_gpio_pad_latch_enable(HW_ID_PORT, HW_ID_BIT_2_PIN);
        if (hw_gpio_get_pin_status(HW_ID_PORT, HW_ID_BIT_2_PIN)) pcb_version = pcb_version | (1<<2);
        hw_gpio_pad_latch_disable(HW_ID_PORT, HW_ID_BIT_2_PIN);

        /* BIT 3 */
        hw_gpio_set_pin_function(HW_ID_PORT, HW_ID_BIT_3_PIN, HW_GPIO_MODE_INPUT, HW_GPIO_FUNC_GPIO);
        hw_gpio_pad_latch_enable(HW_ID_PORT, HW_ID_BIT_3_PIN);
        if (hw_gpio_get_pin_status(HW_ID_PORT, HW_ID_BIT_3_PIN)) pcb_version = pcb_version | (1<<3);
        hw_gpio_pad_latch_disable(HW_ID_PORT, HW_ID_BIT_3_PIN);

        hw_sys_pd_com_disable();
}
