/**
 ****************************************************************************************
 *
 * @file logging_mechanism.c
 *
 * @brief Logging mechanism implementation
 *
 * Copyright (C) 2016-2020 Dialog Semiconductor.
 * This computer program includes Confidential, Proprietary Information
 * of Dialog Semiconductor. All Rights Reserved.
 *
 ****************************************************************************************
 */

#include <stdio.h>
#include "hw_rtc.h"
#include "logging_mechanism.h"

/*********************************************************************
 *
 *       Defines
 *
 *********************************************************************
 */

#define _NEW_LINE_   "\n\r"

/*
 * '1' to enable printing log info on the serial console, otherwise '0'.
 *
 * \note Should be used only for debugging purposes.
 *
 */
#define VERBOSE_STATE     0

/* Some arbitrary magic value for the RTC initialization */
#define LOG_TIMESTAMP_INIT_MAGIC_VALUE   0xAABBCCDD

/*********************************************************************
 *
 *       Macros
 *
 *********************************************************************
 */

/* Generic macro to print text on the serial console */
#define _VERBOSE_(_state, _fmt, _args...)                             \
                                                                      \
        if (_state) {                                                 \
                printf(_NEW_LINE_ #_fmt _NEW_LINE_, ##_args);         \
        }

/*********************************************************************
 *
 *      Retained symbols
 *
 *********************************************************************
 */

/* Magic value used to designate the RTC initialization */
__RETAINED_UNINIT static uint32_t TimestampMagicValue;

/* Array of dump indexers (one for each distance zone) */
__RETAINED_RW static FSIZE_t _aDumpFileIdx[LOG_NUM_OF_DISTANCE_ZONES] = { 0 };

/* Function pointer to timestamp routine (user-defined callback function) */
__RETAINED_RW static get_timestamp_cb pfGetTimestamp = NULL;

/*********************************************************************
 *
 *       Static code
 *
 *********************************************************************
 */

#if !FF_FS_NORTC
/* RTC block initialization (required for timestamp operations) */
static int LOG_InitTimestamp(void)
{
        int Status;

        /* Check if this is the first time the device is powered */
        if (TimestampMagicValue != LOG_TIMESTAMP_INIT_MAGIC_VALUE) {
                Status = hw_rtc_set_time_clndr(&time_cfg, &calendar_cfg);

                if (Status != HW_RTC_VALID_ENTRY) { //! Any invalid entry should cause failure
                        ASSERT_WARNING(0);
                        return LOG_STATUS_INVALID_TIMESTAMP_DEFAULT_VALUES;
                }
                /*
                 * Designate that RTC initialization has been performed
                 * (valid as long as the device is powered).
                 *
                 * \note Alternately, the magic value can be stored in the flash memory used
                 *       so the device is not re-initialized following power failure.
                 */
                TimestampMagicValue = LOG_TIMESTAMP_INIT_MAGIC_VALUE;
        }
        hw_rtc_clock_enable(); //! Make sure the clock gating is activated
        hw_rtc_start(); //! Make sure both time and calendar timers are activated

        _VERBOSE_(VERBOSE_STATE, %s, __func__)
        return LOG_STATUS_OK;
}
#endif

/*********************************************************************
 *
 *       Non-static code
 *
 *********************************************************************
 */

/* Register timestamp callback function */
void LOG_GetTimestampRegisterCB(get_timestamp_cb pfCallback)
{
        if (pfCallback != NULL) {
                pfGetTimestamp = pfCallback;
        }
}

/* Unregister timestamp callback function */
void LOG_GetTimestampUnregisterCB(void)
{
        pfGetTimestamp = NULL;
}

/* LOG mechanism initialization */
int LOG_Init(void)
{
        FIL File;
        FRESULT FileStatus;
        char _acBuff[LOG_MIN_FILENAME_SIZE];

        for (int Idx = 0; Idx < LOG_NUM_OF_DISTANCE_ZONES; Idx++) {
                sprintf(_acBuff, LOG_FILENAME_PATH "%d" LOG_FILENAME_EXTENSION, Idx); //! Get the file name to be accessed

                /*
                 * Just open the file. If the file does not exist, a new one will be created
                 * (required the very first time the device is initiated).
                 *
                 * \note Do not truncate or overwrite the file as we don't know if its contents
                 *       have been dumped by the peer device. This should be done explicitly
                 *       by calling the corresponding API.
                 */
                FileStatus = f_open(&File, _acBuff, FA_READ | FA_WRITE | FA_OPEN_ALWAYS);
                if (FileStatus != FR_OK) { //! Any error should cause failure
                        ASSERT_WARNING(0);
                        return LOG_STATUS_FAIL;
                }
                FileStatus = f_close(&File);
                if (FileStatus != FR_OK) { //! Any error should cause failure
                        ASSERT_WARNING(0);
                        return LOG_STATUS_FAIL;
                }
        }

#if !FF_FS_NORTC
        FileStatus = LOG_InitTimestamp(); //! If enabled, initialize timestamp operations
        if (FileStatus != (FRESULT)LOG_STATUS_OK) {
                ASSERT_WARNING(0);
                return FileStatus;
        }
#endif
        _VERBOSE_(VERBOSE_STATE, %s: %s, __func__, _acBuff)
        return LOG_STATUS_OK;
}

/* Truncate/overwrite all files (related to distance zones) */
int LOG_TruncateFiles(void)
{
        FIL File;
        FRESULT FileStatus;
        int ret = LOG_STATUS_OK;
        char _acBuff[LOG_MIN_FILENAME_SIZE];

        for (int Idx = 0; Idx < LOG_NUM_OF_DISTANCE_ZONES; Idx++) {
                sprintf(_acBuff, LOG_FILENAME_PATH "%d" LOG_FILENAME_EXTENSION, Idx); //! Get the file name to be accessed

                /* Just open the file. If the file exists, it will be truncated and overwritten. */
                FileStatus = f_open(&File, _acBuff, FA_READ | FA_WRITE | FA_CREATE_ALWAYS);
                if (FileStatus != FR_OK) { //! Any error should cause failure
                        ASSERT_WARNING(0);
                        return LOG_STATUS_FAIL;
                }
                if (f_tell(&File)) { //! Verify that file indexer is reset to zero
                        ASSERT_WARNING(0);
                        ret = LOG_STATUS_FAIL;
                        goto exit;
                }

exit:
                FileStatus = f_close(&File);
                if (FileStatus != FR_OK) { //! Any error should cause failure
                        ASSERT_WARNING(0);
                        ret = LOG_STATUS_FAIL;
                }
                _aDumpFileIdx[Idx] = 0; //! Reset dump indexers regardless of the file system's state
        }
        _VERBOSE_(VERBOSE_STATE, %s: %s, __func__, _acBuff);
        return ret;
}

/* Truncate/overwrite a specified file */
int LOG_TruncateFile(int Idx)
{
        FIL File;
        FRESULT FileStatus;
        int ret = LOG_STATUS_OK;
        char _acBuff[LOG_MIN_FILENAME_SIZE];

        if (Idx >= LOG_NUM_OF_DISTANCE_ZONES) { //! Check indexing boundaries (files are zero indexed)
                ASSERT_WARNING(0);
                return LOG_STATUS_FILE_IDX_OUT_OF_BOUNDARIES;
        }
        sprintf(_acBuff, LOG_FILENAME_PATH "%d" LOG_FILENAME_EXTENSION, Idx); //! Get the file name to be accessed

        /* Just open the file. If the file exists, it will be truncated and overwritten. */
        FileStatus = f_open(&File, _acBuff, FA_READ | FA_WRITE | FA_CREATE_ALWAYS);
        if (FileStatus != FR_OK) { //! Any error should cause failure
                ASSERT_WARNING(0);
                return LOG_STATUS_FAIL;
        }
        if (f_tell(&File)) { //! Verify that file indexer is reset
                ASSERT_WARNING(0);
                ret = LOG_STATUS_FAIL;
                goto exit;
        }

exit:
        FileStatus = f_close(&File);
        if (FileStatus != FR_OK) { //! Any error should cause failure
                ASSERT_WARNING(0);
                ret = LOG_STATUS_FAIL;
        }
        _aDumpFileIdx[Idx] = 0; //! Reset dump file indexer regardless of the file system's state

        _VERBOSE_(VERBOSE_STATE, %s: %s, __func__, _acBuff);
        return ret;
}

/* Reset all dump indexers  */
void LOG_ResetDumpIndexers(void)
{
        for (int Idx = 0; Idx < LOG_NUM_OF_DISTANCE_ZONES; Idx++) {
                _aDumpFileIdx[Idx] = 0;
        }
}

/* Reset dump indexer (of a specified file) */
int LOG_ResetDumpIndexer(int Idx)
{
        if (Idx >= LOG_NUM_OF_DISTANCE_ZONES) { //! Check indexing boundaries (files are zero indexed)
                ASSERT_WARNING(0);
                return LOG_STATUS_FILE_IDX_OUT_OF_BOUNDARIES;
        }
        _aDumpFileIdx[Idx] = 0;
        return LOG_STATUS_OK;
}

/* Dump contents of a specified file */
int LOG_DumpFile(int Idx, void * pData, UINT LenRequest, UINT * pLenRead)
{
        FIL File;
        FRESULT FileStatus;
        int ret = LOG_STATUS_OK;
        char _acBuff[LOG_MIN_FILENAME_SIZE];

        if (Idx >= LOG_NUM_OF_DISTANCE_ZONES) { //! Check indexing boundaries (files are zero indexed)
                ASSERT_WARNING(0);
                return LOG_STATUS_FILE_IDX_OUT_OF_BOUNDARIES;
        }
        sprintf(_acBuff, LOG_FILENAME_PATH "%d" LOG_FILENAME_EXTENSION, Idx); //! Get the file name to be accessed

        FileStatus = f_open(&File, _acBuff, FA_READ | FA_OPEN_EXISTING); //! At this point the file should be created
        if (FileStatus != FR_OK) { //! Any error should cause failure
                ASSERT_WARNING(0);
                return LOG_STATUS_FAIL;
        }

        /*
         * Adjust file indexing based on the last dump state.
         *
         * \note The file is opened in read only mode so, any indexing value beyond the actual
         *       file size will not cause the latter to be expanded.
         *
         */
        FileStatus = f_lseek(&File, _aDumpFileIdx[Idx]);
        if (FileStatus != FR_OK) { //! Any error should cause failure
                ASSERT_WARNING(0);
                ret = LOG_STATUS_FAIL;
                goto exit;
        }

        if ( (LenRequest >= 0) && (pData != NULL) ) {
                FileStatus = f_read(&File, pData, LenRequest, pLenRead);

                if (FileStatus != FR_OK) { //! Any error should cause failure
                        ASSERT_WARNING(0);
                        ret = LOG_STATUS_FAIL;
                        goto exit;
                }
                _aDumpFileIdx[Idx] += *pLenRead; //! Update file indexer based on the actual number of data read
        }

exit:
        FileStatus = f_close(&File);
        if (FileStatus != FR_OK) { //! Any error should cause failure
                ASSERT_WARNING(0);
                ret = LOG_STATUS_FAIL;
        }

        _VERBOSE_(VERBOSE_STATE, %s: %s, __func__, _acBuff);
        return ret;
}

/* Log text (null terminated string) into a specified file */
int LOG_WriteText(int Idx, const char * pText)
{
        FIL File;
        FRESULT FileStatus;
        int ret = LOG_STATUS_OK;
        char _acBuff[LOG_MIN_FILENAME_SIZE];

        if (Idx >= LOG_NUM_OF_DISTANCE_ZONES) {  //! Check indexing boundaries (files are zero indexed)
                ASSERT_WARNING(0);
                return LOG_STATUS_FILE_IDX_OUT_OF_BOUNDARIES;
        }
        sprintf(_acBuff, LOG_FILENAME_PATH "%d" LOG_FILENAME_EXTENSION, Idx);

        FileStatus = f_open(&File, _acBuff, FA_WRITE | FA_OPEN_EXISTING | FA_OPEN_APPEND); //! At this point the file should be created
        if (FileStatus != FR_OK) { //! Any error should cause failure
                ASSERT_WARNING(0);
                return LOG_STATUS_FAIL;
        }

        /*
         * No need to flush dirty sectors (f_sync()). The whole text should be written
         * for the log entry to be considered valid.
         */

        if (pfGetTimestamp) { //! Check if there is any timestamp function registered
                FileStatus = f_puts(pfGetTimestamp(), &File);
                if (FileStatus == EOF) {
                        ASSERT_WARNING(0);
                        ret = LOG_STATUS_FILE_OUT_OF_SPACE;
                        goto exit;
                }
        }

        if (pText != NULL) {
                FileStatus = f_puts(pText, &File); //! Write the text given
                if (FileStatus == EOF) {
                        ASSERT_WARNING(0);
                        ret = LOG_STATUS_FILE_OUT_OF_SPACE;
                        goto exit;
                }
        }

exit:
        FileStatus = f_close(&File);
        if (FileStatus != FR_OK) { //! Any error should cause failure
                ASSERT_WARNING(0);
                ret = LOG_STATUS_FAIL;
        }

        _VERBOSE_(VERBOSE_STATE, %s: %s, __func__, _acBuff)
        return ret;
}

/* Log a specified number of data into a specified file */
int LOG_WriteData(int Idx, const void * pData, UINT LenRequest, UINT * pLenRead)
{
        FIL File;
        FRESULT FileStatus;
        int ret = LOG_STATUS_OK;
        char _acBuff[LOG_MIN_FILENAME_SIZE];

        if (Idx >= LOG_NUM_OF_DISTANCE_ZONES) { //! Check indexing boundaries (files are zero indexed)
                ASSERT_WARNING(0);
                return LOG_STATUS_FILE_IDX_OUT_OF_BOUNDARIES;
        }
        sprintf(_acBuff, LOG_FILENAME_PATH "%d" LOG_FILENAME_EXTENSION, Idx); //! Get the file name to be accessed

        FileStatus = f_open(&File, _acBuff, FA_WRITE | FA_OPEN_EXISTING | FA_OPEN_APPEND); //! At this point the file should be created
        if (FileStatus != FR_OK) { //! Any error should cause failure
                ASSERT_WARNING(0);
                return LOG_STATUS_FAIL;
        }

        /*
         * No need to flush dirty sectors (f_sync()). The whole text should be written
         * for the log entry to be considered valid.
         */

        if (pfGetTimestamp) { //!  Check if there is any timestamp function registered
                FileStatus = f_puts(pfGetTimestamp(), &File);
                if (FileStatus == EOF) {
                        ASSERT_WARNING(0);
                        ret = LOG_STATUS_FILE_OUT_OF_SPACE;
                        goto exit;
                }
        }

        /* Log the actual content */
        if ( (LenRequest > 0) && (pData != NULL) ) {
                FileStatus = f_write(&File, pData, LenRequest, pLenRead);

                if (FileStatus != FR_OK) { //! Any error should cause failure
                        ASSERT_WARNING(0);
                        ret = LOG_STATUS_FAIL;
                        goto exit;
                }
        }

exit:
        FileStatus = f_close(&File);
        if (FileStatus != FR_OK) { //! Any error should cause failure
                ASSERT_WARNING(0);
                ret = LOG_STATUS_FAIL;
        }

        _VERBOSE_(VERBOSE_STATE, %s: %s, __func__, _acBuff);
        return ret;
}
