/**
 ****************************************************************************************
 *
 * @file logging_mechanism.h
 *
 * @brief Logging mechanism implementation
 *
 * Copyright (C) 2016-2020 Dialog Semiconductor.
 * This computer program includes Confidential, Proprietary Information
 * of Dialog Semiconductor. All Rights Reserved.
 *
 ****************************************************************************************
 */

#ifndef LOGGING_MECHANISM_H_
#define LOGGING_MECHANISM_H_

#include "ff.h"
#include "commons.h"

/*********************************************************************
 *
 *       Defines
 *
 *********************************************************************
 */

/**
 * \brief Total number of distance zones
 *
 * This value reflects the total number of distance zones supported. A separate file,
 * stored in the flash memory, will be generated for each distance zone and should be
 * accessed to log/dump data related to the zone.
 *
 */
#define LOG_NUM_OF_DISTANCE_ZONES                  ( 4 )

/*
 * \brief File naming
 *
 * This macro reflects the file naming along with the directory path under which
 * a file is stored (related to distance zone). For more info on the supported
 * naming formats refer to the corresponding section.
 *
 * In the current implementation, the files created are zero indexed. In specific,
 * a serial number is appended to the end of a file name so that, it can be easily
 * accessed and distinguished.
 * In doing so, files can be accessed just by providing the unique serial number
 * regardless of the actual file name or directory path.
 *
 * For instance, and given the default macro configurations, the following files
 * will be created in the root directory of logic drive '0':
 *
 * - zone0.txt
 * - zone1.txt
 * - zone2.txt
 * - zone3.txt
 */
#define LOG_FILENAME_PATH               "zone"

/*
 * \brief Filename extension
 *
 * All files (related to distance zones) will be suffixed by this extension.
 *
 */
#define LOG_FILENAME_EXTENSION          ".txt"

/*
 * \brief Minimum file name size
 *
 * This macro reflects the min. number of characters required to display a file name (required by the logging mechanism)
 * This value should be at least: sizeof(LOG_FILENAME_PATH) - 1
 *                                + unique serial number (typically 1-2 characters)
 *                                + sizeof(LOG_FILENAME_EXTENSION) - 1
 *                                + 1 byte for the null termination character
 *
 */
#define LOG_MIN_FILENAME_SIZE        sizeof(LOG_FILENAME_PATH) + sizeof(LOG_FILENAME_EXTENSION) + 2

/**
 * \brief Time display mode
 *
 * '1' to enable the 24H display mode and '0' to enable the 12H display mode.
 *
 * \note The RTC block can be used for timestamp operations both by the logging mechanism
 *       and the file system used (FATFS).
 *       However, the current file system version supports only the 24H display mode and so,
 *       it is advised not to switch to the 12H display mode.
 */
#define RTC_TIME_DISPLAY_MODE_24H       1

/*********************************************************************
 *
 *       Typedefs
 *
 *********************************************************************
 */

/**
 * \brief Log status enumeration
 *
 * Use this macro to get the status of a logging operation.
 */
typedef enum {
        LOG_STATUS_FAIL                               = -1,    //! Requested operation failed to be executed. This is a generic error and should reflect any failure occurred while trying to open/close a file or perform a read/write operation.
        LOG_STATUS_OK,                                         //! Requested operation executed successfully.
        LOG_STATUS_FILE_IDX_OUT_OF_BOUNDARIES,                 //! File indexing is out of boundaries. This error should be returned in case invalid indexes are given to access files.
        LOG_STATUS_FILE_OUT_OF_SPACE,                          //! No more space is available in the flash memory used. This error should be returned in case the file system has run out of memory. Note that this error does not cover all use cases. That is, running out of memory can also be indicated by returning the generic error!
        LOG_STATUS_INVALID_TIMESTAMP_DEFAULT_VALUES            //! RTC invalid default values. This error should be returned in case invalid values are given to the RTC configuration macros e.g. RTC_TIME_SECOND etc.
} LOG_STATUS;

/**
 * User-defined callback function to get timestamp
 *
 * \warning The timestamp value returned should be provided in a text format
 *          and it's essential that it contain a null termination character!
 *
 */
typedef const char * (* get_timestamp_cb)(void);

/*********************************************************************
 *
 *       APIs
 *
 *********************************************************************
 */

/**
 * \brief Logging mechanism initialization
 *
 * This routine should be called prior to performing any log/dump operations.
 * A specified number of files are created and the RTC block is initialized
 * if 'FF_FS_NORTC' is set to '0'.
 *
 * \return The status of the target operation
 *
 * \sa LOG_STATUS_OK
 * \sa LOG_STATUS_FAIL
 *
 */
int LOG_Init(void);

/**
 * \brief Truncate all files
 *
 * This is a helper function that should be used to truncate all files (related to
 * distance zones). In doing so, the internal file pointers are reset and so files
 * can be overwritten with new contents. File indexers, used for dump operations,
 * are also reset.
 *
 * \note This routine should be called periodically, e.g. after dumping file contents,
 *       so that the file system does not run out of memory.
 *
 * \return The status of the target operation
 *
 * \sa LOG_STATUS_OK
 * \sa LOG_STATUS_FAIL
 *
 */
int LOG_TruncateFiles(void);

/**
 * \brief Truncate file
 *
 * This routine should be used to truncate a specified file (related to distance zone).
 * In doing so, the internal file pointer is reset and so the file can be overwritten
 * with new contents. The corresponding dump indexer is also reset.
 *
 * \param[in]  Idx  File indexer to be truncated
 *
 * \note This routine should be called periodically, e.g. after dumping file contents,
 *       so that the file system does not run out of memory.
 *
 * \return The status of the target operation
 *
 * \sa LOG_STATUS_OK
 * \sa LOG_STATUS_FAIL
 * \sa LOG_STATUS_FILE_IDX_OUT_OF_BOUNDARIES
 *
 */
int LOG_TruncateFile(int Idx);

/**
 * \brief Reset all dump indexers
 *
 * This is a helper function that should be used to reset all dump indexers
 * (associated to distance zones). For more info refer to \sa 'LOG_ResetDumpIndexers()'
 *
 */
void LOG_ResetDumpIndexers(void);

/**
 * \brief Reset dump indexer
 *
 * The logging mechanism keeps track of, and for each file associated to distance zone, the last
 * chunk of data dumped (to the peer device). This is achieved by just tracking the internal file
 * pointer. The dump indexer advances as long as new chunks of data are dumped in to the file.
 * Once the End Of File (EOF) is reached, the dump indexer is no longer advancing.
 *
 * \note This routine can be called once the End Of File (EOF) is reached. In doing so, the file can
 *       be dumped from the beginning.
 *
 * \param[in]  Idx  Dump indexer to be reset
 *
 * \return The status of the target operation
 *
 * \sa LOG_STATUS_OK
 * \sa LOG_STATUS_FILE_IDX_OUT_OF_BOUNDARIES
 *
 */
int LOG_ResetDumpIndexer(int Idx);

/**
 * \brief Log data
 *
 * This routine should be used to write/log a specified number of bytes to a specified file.
 *
 * \param[in]    Idx            File indexer to be accessed
 * \param[in]    pData          Pointer to buffer containing the data to be logged (in the flash memory used)
 * \param[in]    LenRequest     Number of data requested to be logged (by the target application)
 * \param[out]   pLenRead       Number of data actually logged (by the file system)
 *
 * \return The status of the target operation
 *
 * \note Invalid file indexers will cause the routine to return immediately
 *
 * \note A zero value of \p pLenRead designates end of memory. That is, no more memory space
 *       is available to be used by the file system.
 *
 * \note If timestamp callback is registered, some timestamp text will be written/logged before
 *       writing the actual data.
 *
 * \sa LOG_STATUS_OK
 * \sa LOG_STATUS_FAIL
 * \sa LOG_STATUS_FILE_OUT_OF_SPACE
 * \sa LOG_STATUS_FILE_IDX_OUT_OF_BOUNDARIES
 *
 */
int LOG_WriteData(int Idx, const void * pData, UINT LenRequest, UINT * pLenRead);

/**
 * \brief Log text
 *
 * This routine should be used to write/log text to a specified file. Once the termination
 * character is found, and without writing it, the logging operation will be stopped.
 *
 * \param[in] Idx      File indexer to be accessed
 * \param[in] pText    Pointer to buffer containing the text to be logged (in the flash memory used)
 *
 * \return The status of the target operation
 *
 * \note Invalid file indexers will cause the routine to return immediately
 *
 * \note If timestamp callback is registered, some timestamp text will be written/logged before
 *       writing the actual data.
 *
 * \warning The text should contain a null termination character
 *
 * \sa LOG_STATUS_OK
 * \sa LOG_STATUS_FAIL
 * \sa LOG_STATUS_FILE_OUT_OF_SPACE
 * \sa LOG_STATUS_FILE_IDX_OUT_OF_BOUNDARIES
 *
 */
int LOG_WriteText(int Idx, const char * pText);

/**
 * \brief Dump data (of a specified file)
 *
 * \param[in]    Idx           File indexer to be accessed
 * \param[in]    pData         Pointer to buffer in which dumped data will be stored
 * \param[in]    LenRequest    Number of data requested to be dumped (by the target application)
 * \param[out]   pLenRead      Number of data actually dumped (by the file system)
 *
 * \note A zero value of \p pLenRead designates end of dumping. That is, all file contents
 *       have been read/dumped. It's a prerequisite that dump operations are executed with
 *       no errors.
 *
 * \note Invalid file indexers will cause the routine to return immediately
 *
 * \return The status of the target operation
 *
 * \sa LOG_STATUS_OK
 * \sa LOG_STATUS_FAIL
 * \sa LOG_STATUS_FILE_IDX_OUT_OF_BOUNDARIES
 *
 */
int LOG_DumpFile(int Idx, void * pData, UINT LenRequest, UINT * pLenRead);

/**
 * \brief Get timestamp callback function registration
 *
 * \param[in]  cb  Callback function to be registered (to the logging mechanism)
 *
 * By default, no timestamp operations take place. To do so, a callback routine
 * should be explicitly registered (by the application). The routine registered
 * will be called prior to performing logging operations.
 *
 * \warning It is essential that the pointer returned by the callback points to
 *          a null terminated text.
 *
 */
void LOG_GetTimestampRegisterCB(get_timestamp_cb pfCallback);

/**
 * \brief Get timestamp callback function un-registration
 *
 * This routine should be called to unregister a previously defined get timestamp
 * callback function. In doing so, no timestamp operations will be performed
 * following logging operations.
 *
 */
void LOG_GetTimestampUnregisterCB(void);

#endif /* LOGGING_MECHANISM_H_ */
