/**
 ****************************************************************************************
 *
 * @file miscellaneous.h
 *
 * @brief Miscellaneous functionality
 *
 * Copyright (C) 2016-2020 Dialog Semiconductor.
 * This computer program includes Confidential, Proprietary Information
 * of Dialog Semiconductor. All Rights Reserved.
 *
 ****************************************************************************************
 */

#include <stdio.h>
#include "ff.h"
#include "sys_timer.h"
#include "hw_rtc.h"
#include "logging_mechanism.h"
#include "miscellaneous.h"

/*********************************************************************
 *
 *       Defines
 *
 *********************************************************************
 */

#define GET_TIMESTAMP_MULTIPLIER          1000000UL
#define GET_TIMESTAMP_DIVIDER             configSYSTICK_CLOCK_HZ

#define _NEW_LINE_       "\n\r"

/* Enable/disable printing debugging info on the serial console. */
#define VERBOSE_STATE      1

/* Minimum size required to generate the text to be logged */
#define LOG_ENTRY_MIN_TEXT_SIZE  ( sizeof(LOG_TEXT_ENTRY) + sizeof(DISTANCE_ZONE_BLUETOOTH_LIMIT) + 27)


/*********************************************************************
 *
 *       Macros
 *
 *********************************************************************
 */

/* Generic macro to print log info on the serial console */
#define _VERBOSE_(_state, _fmt, _args...)                             \
                                                                      \
        if (_state) {                                                 \
                printf(_NEW_LINE_ #_fmt _NEW_LINE_, ##_args);         \
        }

/*********************************************************************
 *
 *       Static code
 *
 *********************************************************************
 */

static const char * _acDistanceZone[] = {
        DISTANCE_ZONE_DANGER,
        DISTANCE_ZONE_WARNING,
        DISTANCE_ZONE_COARSE,
        DISTANCE_ZONE_BLUETOOTH_LIMIT,
};

/**
 * Number of low power (lp) clock ticks upon device's boot.
 *
 * This routine returned the number of lp clock ticks elapsed since
 * the device booted.
 *
 * \note Time can also be retrieved directly by calling the following APIs:
 *
 *       \sa sys_timer_get_uptime_usec()
 *       \sa sys_timer_get_uptime_usec_fromISR()
 *
 */
static uint64_t get_sys_timestamp(void)
{
        return (in_interrupt()) ? sys_timer_get_uptime_ticks_fromISR() :
                                                        sys_timer_get_uptime_ticks();
}

/**
 * \brief Get file information
 *
 * \param[in]   Idx    File indexer to be requested
 * \param[out]  pInfo  Pointer to info structure as required by the file system
 *
 * \return The status of the target operation
 *
 * \sa LOG_STATUS_FILE_IDX_OUT_OF_BOUNDARIES
 *
 */
static int LOG_GetFileInfo(int Idx, FILINFO * pInfo)
{
        FRESULT FileStatus;
        char _acBuff[LOG_MIN_FILENAME_SIZE];

        if (Idx >= LOG_NUM_OF_DISTANCE_ZONES) { //! First check the boundaries (boundaries are zero indexed)
                ASSERT_WARNING(0);
                return LOG_STATUS_FILE_IDX_OUT_OF_BOUNDARIES;
        }
        sprintf(_acBuff, LOG_FILENAME_PATH "%d" LOG_FILENAME_EXTENSION, Idx); //! Get the file name to be accessed

        FileStatus = f_stat(_acBuff, pInfo);
        return FileStatus;
}

/*********************************************************************
 *
 *       Non static code
 *
 *********************************************************************
 */

/* FATFS initialization */
int FATFS_Init(void)
{
        FRESULT FileStatus;

        /* Initialize the FATFS */
        FileStatus = f_init();

        if (FileStatus != FR_OK) { //! Any error should cause failure
                ASSERT_WARNING(0);
                return -1;
        }
        return 0;
}

/* Get time duration is millisecond */
void LOG_MeasureTimeInMS(GET_TIMESTAMP State, uint32_t * Integer, uint16_t * Fractional)
{
        static uint64_t TimestampStart;
        uint64_t TimestampStop;
        uint64_t usTime;

        if (State == GET_TIMESTAMP_START) {
                TimestampStart = get_sys_timestamp();
        }
        else if (State == GET_TIMESTAMP_STOP) {
                TimestampStop = get_sys_timestamp();
                usTime = ((TimestampStop - TimestampStart) * GET_TIMESTAMP_MULTIPLIER / GET_TIMESTAMP_DIVIDER);

                if (Integer != NULL) {
                        *Integer = (uint32_t)(usTime / 1000);
                }
                if (Fractional != NULL) {
                        *Fractional = (uint16_t)((usTime / 100) % 10);
                }
                _VERBOSE_(VERBOSE_STATE, Transfer: %lu.%d ms, (uint32_t)(usTime / 1000),
                                                                           (uint16_t)(usTime / 100) % 10)
        }
}

/* Get the current date/time from the RTC block */
const char * LOG_GetTimestampCallback(void)
{
        static char _acTimestamp[LOG_TIMESTAMP_MIN_TEXT_SIZE];
        rtc_time time_info;
        rtc_calendar calendar_info;

        /* Get the current time/date */
        hw_rtc_get_time_clndr(&time_info, &calendar_info);

        sprintf(_acTimestamp, LOG_TIMESTAMP_ENTRY, calendar_info.wday, calendar_info.mday, calendar_info.month,
                                                                                            calendar_info.year,
                                                              time_info.hour, time_info.minute, time_info.sec);
        return _acTimestamp;

}

/* Prepare some text to be logged */
const char * LOG_PrepareLogEntry(const char * pDeviceAddress, LOG_DISTANCE_ZONE DistanceZone)
{
        static char Buff[LOG_ENTRY_MIN_TEXT_SIZE];

        sprintf(Buff, LOG_TEXT_ENTRY, pDeviceAddress, _acDistanceZone[DistanceZone]);
        return Buff;
}

/* Set RTC counters based on the date/time a specified file was accessed.  */
int LOG_SetRTCBasedOnLastFileAccess(int Idx)
{
        FILINFO FileInfo;
        FRESULT FileStatus;
        DWORD FileTimestamp;
        rtc_time time_info;
        rtc_calendar clndr_info;

        /* Get file info */
        FileStatus = LOG_GetFileInfo(Idx, &FileInfo);
        if (FileStatus != FR_OK) {
                ASSERT_WARNING(0);
                return LOG_STATUS_FAIL;
        }
        /* Unify the time/data info retrieved from the FS. */
        FileTimestamp = FS_TIME_PREPARE(FileInfo.ftime, FileInfo.fdate);

        /* Unpack the timestamp info retrieved from the FS. */
        FS_TIME_UNPACK(time_info.sec, time_info.minute, time_info.hour, clndr_info.mday,
                                                        clndr_info.month, clndr_info.year, FileTimestamp);

        /* Make sure the rest RTC values are initialized correctly */
        time_info.hour_mode = RTC_TIME_DISPLAY_MODE_24H ? RTC_24H_CLK : RTC_12H_CLK;
        time_info.hsec = 0; //! Don't care
        clndr_info.wday = RTC_CALENDAR_WEEK_DAY; //! Note: this value cannot be requested by the file system, use the corresponding configuration macro.

        FileStatus = hw_rtc_set_time_clndr(&time_info, &clndr_info);
        if (FileStatus != (FRESULT)HW_RTC_VALID_ENTRY) { //! Any invalid entry should cause failure
                ASSERT_WARNING(0);
                return LOG_STATUS_INVALID_TIMESTAMP_DEFAULT_VALUES;
        }

        _VERBOSE_(VERBOSE_STATE, %s: File modified on %02d:%02d:%02d (time)  %02d-%02d-%02d (date), __func__,
                                                                               time_info.hour, time_info.minute,
                                                                               time_info.sec,
                                                                               clndr_info.mday, clndr_info.month,
                                                                               clndr_info.year)
        return LOG_STATUS_OK;
}

/* Get the device address in text format */
const char * FormatDeviceAddress(const bd_address_t * pDeviceAddress)
{
        static char Buf[19];
        int i;

        for (i = 0; i < sizeof(pDeviceAddress->addr); i++) {
                int idx;

                idx = sizeof(pDeviceAddress->addr) - i - 1;
                sprintf(&Buf[i * 3], "%02x:", pDeviceAddress->addr[idx]);
        }
        Buf[sizeof(Buf) - 2] = '\0';

        return Buf;
}
