/**
 ****************************************************************************************
 *
 * @file miscellaneous.h
 *
 * @brief Miscellaneous functionality
 *
 * Copyright (C) 2016-2020 Dialog Semiconductor.
 * This computer program includes Confidential, Proprietary Information
 * of Dialog Semiconductor. All Rights Reserved.
 *
 ****************************************************************************************
 */

#ifndef MISCELLANEOUS_MISCELLANEOUS_H_
#define MISCELLANEOUS_MISCELLANEOUS_H_

#include "logging_mechanism.h"
#include "ble_common.h"

/*********************************************************************
 *
 *       Defines
 *
 *********************************************************************
 */


/*
 * Chunk of data used to construct the timestamp content. The text adopts an XML style
 * so that info can be easily parsed (by the peer device).
 */
#define LOG_TIMESTAMP_STARTING_TAG   "\n\r<timestamp>"
#define LOG_TIMESTAMP_ENDING_TAG      "</timestamp>\n\r"

/*
 * Recommended timestamp content (in XML format)
 *
 * \warning Any format can be adopted. However, if the RTC counter values (%02d) are changed
 *          \sa LOG_GetTimestampCallback() should be modified accordingly.
 *
 */
#define LOG_TIMESTAMP_DATA         "\n\r   <week>%02d</week>\n\r   <date>%02d-%02d-%02d</date>\n\r   <time>%02d:%02d:%02d</time>\n\r"

/*
 * Recommended timestamp (in XML format) to be used in timestamp operations
 *
 * \sa LOG_GetTimestampCallback()
 *
 */
#define LOG_TIMESTAMP_ENTRY   ( LOG_TIMESTAMP_STARTING_TAG LOG_TIMESTAMP_DATA LOG_TIMESTAMP_ENDING_TAG )

/* Minimum timestamp context length required by the corresponding API. */
#define LOG_TIMESTAMP_MIN_TEXT_SIZE       sizeof(LOG_TIMESTAMP_ENTRY)

/*
 * Chunk of data used to construct the XML text to be logged. The text adopts an XML style
 * so that info can be easily parsed (by the peer device).
 */
#define LOG_TEXT_STARTING_TAG     "\n\r<log>"
#define LOG_TEXT_ENDING_TAG       "</log>\n\r"
#define LOG_TEXT_DATA             "\n\r   <address>%s</address>\n\r   <zone>%s</zone>\n\r"

/* Recommended text (in XML format) to be logged */
#define LOG_TEXT_ENTRY            ( LOG_TEXT_STARTING_TAG LOG_TEXT_DATA LOG_TEXT_ENDING_TAG )

/* Distance zones in text format (used by \sa LOG_PrepareLogEntry()) */
#define DISTANCE_ZONE_BLUETOOTH_LIMIT   "Bluetooth Limit"
#define DISTANCE_ZONE_COARSE            "Coarse Ranging"
#define DISTANCE_ZONE_WARNING           "Warning Ranging"
#define DISTANCE_ZONE_DANGER            "Danger Zone"


/*********************************************************************
 *
 *       Typedefs
 *
 *********************************************************************
 */

typedef enum {
        GET_TIMESTAMP_START  = 0x0,
        GET_TIMESTAMP_STOP
} GET_TIMESTAMP;

/*********************************************************************
 *
 *       APIs
 *
 *********************************************************************
 */

/**
 * \brief FATFS Initialization
 *
 * This is a wrapper function that should be called to initialize the FATFS.
 * The routine is recommended to be called in \sa retarget_init().
 *
 * \warning Make sure the routine is called prior to performing any file system operations.
 *
 */
int FATFS_Init(void);

/**
 * \brief Measure time (ms)
 *
 * This routine should be used to measure time in milliseconds. The API can be used for
 * measuring time execution (of functions) or for timestamping logging operations.
 * (instead of employing the RTC block).
 *
 * \param[in]   State         The state indicating the start/stop of a measurement cycle.
 * \param[out]  Integer       The integer part of the value measured
 * \param[out]  Fractional    The fractional part of the value measured
 *
 * Example usage:
 *
 * \code{.c}
 *
 *      uint32_t TimestampInteger;
 *      uint16_t TimestampFractional;
 *
 *      LOG_MeasureTimeMS(GET_TIMESTAMP_START, &TimestampInteger, &TimestampFractional)
 *
 *      Some code block...
 *
 *      LOG_MeasureTimeMS(GET_TIMESTAMP_START_STOP, &TimestampInteger, &TimestampFractional)
 *
 *      printf("\n\rTime execution (ms): %lu : %d\n\r", TimestampInteger, TimestampFractional);
 *
 * \endcode
 *
 */
void LOG_MeasureTimeInMS(GET_TIMESTAMP State, uint32_t * Integer, uint16_t * Fractional);

/**
 * \brief Timestamp callback function
 *
 * Helper function that can be used for timestamping operations. If registered, timestamp
 * operations will take place prior to logging/writing the actual data. This routine gets
 * the current date/time from the RTC block.
 *
 * \return Pointer to a static buffer containing the timestamp (in text format) to be logged.
 *         The buffer contains a null termination character thus, indicating the end of text.
 *
 * \note It's a prerequisite that the RTC block is up and running before calling the routine
 *
 * \note The timestamp content adopts an XML format which is commonly used by logging systems
 *       and can be easily parsed by a gateway (e.g. raspberry pi). The actual text content
 *       is declared via \sa LOG_TIMESTAMP_DATA
 *
 * Example usage:
 *
 * \code{.c}
 *      // Register callback to get timestamp values
 *      LOG_GetTimestampRegisterCB(LOG_GetTimestampCallback);
 * \endcode
 *
 */
const char * LOG_GetTimestampCallback(void);

/**
 * \brief Format BD address
 *
 * Helper function that can be used to format the device address from numerical value to text.
 * The device address structure is the one used by the BLE adapter implementation.
 *
 * \param[in]  pDeviceAddress  Pointer to BD address structure
 *
 * \return Pointer to static buffer containing the device address in text format. The buffer contains
 *         a null termination character thus, indicating the end of text.
 *
 * \sa LOG_PrepareLogEntry()
 *
 */
const char * FormatDeviceAddress(const bd_address_t * pDeviceAddress);

/**
 * \brief Generate text to be logged
 *
 * Helper function that can be used to generate some text to be logged. The message generated
 * consists of a device address (BD) along with distance zone the device is detected.
 *
 * \param[in]  pDeviceAddress  Pointer to buffer containing the device address in text format
 * \param[in]  DistanceZone    A valid enumeration value associated to a distance zone
 *
 * \return Pointer to a static buffer containing the text to be logged. The buffer contains
 *         a null termination character thus, indicating the end of text.
 *
 * \note The text generated adopts an XML format which is commonly used by logging systems
 *        and can be easily parsed by a gateway (e.g. raspberry pi). The actual text
 *        content is declared via \sa LOG_TEXT_ENTRY
 *
 * Example usage:
 *
 * \code{.c}
 *
 *      int ret;
 *      int Idx;
 *
 *      bd_address_t DeviceAddress = {
 *              .addr = {0x65, 0xBF, 0xC5, 0x60, 0xDF, 0x01}
 *      };
 *      Idx = 0;
 *
 *      ret = LOG_WriteText(Idx, LOG_PrepareLogEntry(FormatDeviceAddress(&DeviceAddress),
 *                                                                      LOG_DISTANCE_ZONE_DANGER));
 *      ASSERT_WARNING(ret == LOG_STATUS_OK);
 *
 * \endcode
 *
 * \sa LOG_WriteText()
 * \sa FormatDeviceAddress()
 *
 */
const char * LOG_PrepareLogEntry(const char * pDeviceAddress, LOG_DISTANCE_ZONE DistanceZone);

/**
 * \brief Configure RTC counters
 *
 * Helper function that can be used to configure the RTC date/time counter values based on the
 * info retrieved from specified file. This info reflects the last date/time the file was
 * modified/accessed.
 * The file requested should already exist. Otherwise, the routine will exit with generic error code.
 *
 * \param[in]  Idx  File index to be requested
 *
 * \return The status of the target operation
 *
 * \note This routine can be employed in case the battery is completely drained and the device
 *       was turned off and no Current Time Service (CTS) is utilized (by the application).
 *
 * \warning The week day info is not used by the file system and so, this value
 *          is initialized via \sa RTC_CALENDAR_WEEK_DAY
 *
 *
 * \sa LOG_STATUS_OK
 * \sa LOG_STATUS_FAIL
 * \sa LOG_STATUS_FILE_IDX_OUT_OF_BOUNDARIES
 * \sa LOG_STATUS_INVALID_TIMESTAMP_DEFAULT_VALUES
 *
 */
int LOG_SetRTCBasedOnLastFileAccess(int Idx);

#endif /* MISCELLANEOUS_MISCELLANEOUS_H_ */
